from django.utils.translation import gettext_lazy as _
from django import forms
from django.utils.safestring import mark_safe

from reface.dashboard import fields
from resh.domains import dali1, dali2
from reface import const as v


class _LineUtils(forms.Form):
    DALI1 = 'dali'
    DALI2 = 'dali2'
    ALL = 'all'

    ACTION = 'action'
    START = 'start'
    STOP = 'stop'
    SCAN = 'scan'

    EXTEND = 'extend_line'
    CONSTRUCT = 'construct_line'
    DOUBLE_ADDRESSES = 'fix_double_addresses'

    VERBOSE_EXTEND = _('Extend line')
    VERBOSE_CONSTRUCT = _('Construct line')
    VERBOSE_DOUBLE_ADDRESSES = _('Fix double addresses')

    COMMANDS = {
        DALI1: {
            EXTEND: dali1.extend_line,
            CONSTRUCT: dali1.construct_line,
            DOUBLE_ADDRESSES: dali1.fix_double_addresses,
            SCAN: dali1.dump,
        },
        DALI2: {
            EXTEND: dali2.extend_line,
            CONSTRUCT: dali2.construct_line,
            DOUBLE_ADDRESSES: dali2.fix_double_addresses,
            SCAN: dali2.dump,
        }
    }

    DALI_TYPES = (('', ''), (DALI1, 'DALI 1'), (DALI2, 'DALI 2'))


class LineBuilder(_LineUtils):
    dali_type = forms.ChoiceField(choices=_LineUtils.DALI_TYPES, label=_('Device types'))
    cross = forms.BooleanField(initial=True, required=False, label=_('Cross addresses'))
    first_address = forms.IntegerField(initial=0, min_value=0, max_value=64, label=_('Start from'))

    @classmethod
    def verbose_name(cls, command_name):
        return {
            cls.EXTEND: cls.VERBOSE_EXTEND,
            cls.CONSTRUCT: cls.VERBOSE_CONSTRUCT,
            cls.DOUBLE_ADDRESSES: cls.VERBOSE_DOUBLE_ADDRESSES,
        }.get(command_name, _('Device addressing'))

    @property
    def mode(self): return 'CROSS' if self.cleaned_data['cross'] else 'NOCROSS'

    def command(self, action): return self.COMMANDS[self.cleaned_data['dali_type']][action]

    @property
    def start_address(self): return self.cleaned_data.get('first_address', 0)


class LineScanner(_LineUtils):
    @classmethod
    def columns(cls):
        return {
            v.COMMON: [dict(name=name, verbose=verb, desc=desc, select=select) for name, verb, desc, select in (
                (v.GTIN, 'Gtin', _('GtinDescription'), None),
                (v.SERIAL, 'Serial', _('SerialDescription'), None),
                (v.GTIN_OEM, 'GtinOem', _('GtinOemDescription'), None),
                (v.SERIAL_OEM, 'SerialOem', _('SerialOemDescription'), None),
                (v.FIRMWARE, 'F.Ver', _('F.VerDescription'), None),
                (v.HARDWARE, 'H.Ver', _('H.VerDescription'), None),
                (v.GROUPS, 'Groups', _('GroupsDescription'), None),
                (v.OPERATING_MODE, 'Op.Mode', _('Op.ModeDescription'), v.RANGE_OPERATING_MODE),
            )],
            v.DALI: [dict(name=name, verbose=verb, desc=desc, select=select) for name, verb, desc, select in (
                (v.FADE_TIME, 'FadeTime', _('FadeTimeDescription'), v.RANGE_FADE_TIME),
                (v.FADE_RATE, 'FadeRate', _('FadeRateDescription'), v.RANGE_FADE_RATE),
                (v.DIMMING_CURVE, 'DimmingCurve', _('DimmingCurveDescription'), v.RANGE_DIMMING_CURVE),
                (v.ACTUAL_LEVEL, 'Level', _('LevelDescription'), v.RANGE_0_255_PERCENT),
                (v.MIN_LEVEL, 'MinLevel', _('MinLevelDescription'), v.RANGE_0_255_PERCENT),
                (v.MAX_LEVEL, 'MaxLevel', _('MaxLevelDescription'), v.RANGE_0_255_PERCENT),
                (v.PHYSICAL_MINIMUM, 'Ph.Min', _('Ph.MinDescription'), None),
                (v.POWER_ON_LEVEL, 'PowerOn', _('PowerOnDescription'), v.RANGE_0_255_PERCENT),
                (v.SYSTEM_FAILURE, 'FailLevel', _('FailLevelDescription'), v.RANGE_0_255),
                (v.SCENES, 'Scenes', _('ScenesDescription'), None),
            )],
            v.SWITCH: [dict(name=name, verbose=verb, desc=desc, select=select) for name, verb, desc, select in (
                (v.SWITCH_UP_ON, 'UpOn', _('UpOnDescription'), v.RANGE_0_255),
                (v.SWITCH_UP_OFF, 'UpOff', _('UpOffDescription'), v.RANGE_0_255),
                (v.SWITCH_DOWN_ON, 'DownOn', _('DownOnDescription'), v.RANGE_0_255),
                (v.SWITCH_DOWN_OFF, 'DownOff', _('DownOffDescription'), v.RANGE_0_255),
                (v.VIRTUAL_LEVEL, 'V.Level', _('V.LevelDescription'), None),
            )],
            v.COLOUR: [dict(name=name, verbose=verb, desc=desc, select=select) for name, verb, desc, select in (
                (v.CC_COOLEST, 'Coolest', _('CoolestDescription'), None),
                (v.CC_WARMEST, 'Warmest', _('WarmestDescription'), None),
                (v.CC_PH_COOLEST, 'Ph.Coolest', _('Ph.CoolestDescription'), None),
                (v.CC_PH_WARMEST, 'Ph.Warmest', _('Ph.WarmestDescription'), None),
                (v.CC_COLOR_TEMPERATURE, 'Temperature', _('TemperatureDescription'), None),
            )],
            v.DALI2: [dict(name=name, verbose=verb, desc=desc, select=select) for name, verb, desc, select in (
                (v.IS_ENABLED, 'Enabled', _('EnabledDescription'), None),
                (v.EVENT_SCHEME, 'EventScheme', _('EventSchemeDescription'), v.RANGE_EVENT_SCHEME),
                (v.EVENT_FILTER, 'EventFilter', _('EventFilterDescription'), v.RANGE_EVENT_FILTER),
                (v.INSTANCE_GROUP_0, 'Group0', _('Group0Description'), v.RANGE_0_31_GROUP),
                (v.INSTANCE_GROUP_1, 'Group1', _('Group1Description'), v.RANGE_0_31_GROUP),
                (v.INSTANCE_GROUP_2, 'Group2', _('Group2Description'), v.RANGE_0_31_GROUP),
                (v.EVENT_PRIORITY, 'Priority', _('PriorityDescription'), v.RANGE_2_5),
                (v.RESOLUTION, 'Res.', _('ResDescription'), None),
                (v.FEATURE_TYPES, 'Features', _('FeaturesDescription'), None),
            )],
            v.SENSORS: [dict(name=name, verbose=verb, desc=desc, select=select) for name, verb, desc, select in (
                (v.DEAD_TIME, 'DeadTime', _('DeadTimeDescription'), v.TIMER_05_0_255_2),
                (v.REPORT_TIME, 'ReportTime', _('ReportTimeDescription'), v.TIMER_1_0_255),
                (v.HYSTERESIS, 'Hysteresis', _('HysteresisDescription'), v.RANGE_0_25),
                (v.HYSTERESIS_MIN, 'Hys.Min', _('Hys.MinDescription'), v.RANGE_0_25),
                (v.CATCHING, 'Catching', _('CatchingDescription'), None),
                (v.HOLD_TIME, 'HoldTime', _('HoldTimeDescription'), v.TIMER_10_0_254),
            )],
            v.BUTTONS: [dict(name=name, verbose=verb, desc=desc, select=select) for name, verb, desc, select in (
                (v.SHORT_TIME, 'ShortTime', _('ShortTimeDescription'), v.TIMER_02_10_255_2),
                (v.SHORT_TIME_MIN, 'ShortMin', _('ShortMinDescription'), v.TIMER_02_10_255_2),
                (v.DOUBLE_TIME, 'DoubleTime', _('DoubleTimeDescription'), v.TIMER_02_10_100_2),
                (v.DOUBLE_TIME_MIN, 'DoubleMin', _('DoubleMinDescription'), v.TIMER_02_10_100_2_MIN),
                (v.REPEAT_TIME, 'RepeatTime', _('RepeatTimeDescription'), v.TIMER_02_5_100_2),
                (v.STUCK_TIME, 'StuckTime', _('StuckTimeDescription'), v.TIMER_1_5_255),
            )],
            v.EMERGENCY: [dict(name=name, verbose=verb, desc=desc, select=select) for name, verb, desc, select in (
                (v.EMERGENCY_MODE, 'EmergencyMode', _('Emg.ModeDescription'), None),
                (v.EMERGENCY_CHARGE_LEVEL, 'ChargeLevel', _('ChargeLevelDescription'), None),
                (v.EMERGENCY_STATUS, 'EmergencyStatus', _('Emg.StatusDescription'), None),
                (v.EMERGENCY_FEATURES, 'EmergencyFeatures', _('Emg.FeaturesDescription'), None),
                (v.EMERGENCY_FAILURE_STATUS, 'FailStatus', _('FailStatusDescription'), None),
                (v.EMERGENCY_DURATION_TEST_RESULT, 'DurationResult', _('DurationResultDescription'), None),
                (v.EMERGENCY_LEVEL, 'EmergencyLevel', _('Emg.LevelDescription'), None),
                (v.EMERGENCY_MAX_LEVEL, 'EmergencyMaxLevel', _('Emg.MaxLevelDescription'), None),
                (v.EMERGENCY_MIN_LEVEL, 'EmergencyMinLevel', _('Emg.MinLevelDescription'), None),
                # (v.EMERGENCY_TEST_TIMING, 'TestTiming', _('TestTimingDescription'), None),
                (v.EMERGENCY_TEST_TIMING_1, 'UntilNextFunctionTest', None, None),
                (v.EMERGENCY_TEST_TIMING_2, 'UntilNextDurationTest', None, None),
                (v.EMERGENCY_TEST_TIMING_3, 'FunctionTestInterval', None, None),
                (v.EMERGENCY_TEST_TIMING_4, 'DurationTestInterval', None, None),
                (v.EMERGENCY_TEST_TIMING_5, 'TestExecutionTimeoutDays', None, None),
                (v.EMERGENCY_TEST_TIMING_6, 'ProlongTimeSeconds', None, None),
            )],
        }

    @classmethod
    def commands(cls):
        return {
            v.COMMON: [
                dict(name=name, verbose=verb, desc=desc) for name, verb, desc in
                (
                    (dali1.reset.__name__, 'Reset', _('ResetDescription')),
                    (dali1.identify.__name__, 'Identify', _('IdentifyDescription')),
                )],
            v.DALI: [
                dict(name=name, verbose=verb, desc=desc) for name, verb, desc in
                (
                    (dali1.discovery.__name__, 'Discovery', _('DiscoveryDescription')),
                    (dali1.off.__name__, 'Off', _('OffDescription')),
                    (dali1.up.__name__, 'Up', _('UpDescription')),
                    (dali1.down.__name__, 'Down', _('DownDescription')),
                    (dali1.step_up.__name__, 'StepUp', _('StepUpDescription')),
                    (dali1.step_down.__name__, 'StepDown', _('StepDownDescription')),
                    (dali1.recall_min_level.__name__, 'MinLevel', _('OnMinLevelDescription')),
                    (dali1.recall_max_level.__name__, 'MaxLevel', _('OnMaxLevelDescription')),
                    (dali1.step_down_and_off.__name__, 'StepDownOff', _('StepDownOffDescription')),
                    (dali1.on_and_step_up.__name__, 'OnStepUp', _('OnStepUpDescription')),
                    (dali1.go_to_last_active_level.__name__, 'LastLevel', _('LastLevelDescription')),
                )],
            v.EMERGENCY: [
                dict(name=name, verbose=verb, desc=desc) for name, verb, desc in
                (
                    (dali1.start_emergency_function_test.__name__, 'FunctionTest', _('FunctionTestDescription')),
                    (dali1.start_emergency_duration_test.__name__, 'DurationTest', _('DurationTestDescription')),
                    (dali1.stop_emergency_test.__name__, 'StopTest', _('StopTestDescription')),
                    (dali1.emergency_inhibit.__name__, 'Inhibit', _('InhibitDescription')),
                    (dali1.emergency_reset_inhibit.__name__, 'ResetInhibit', _('ResetInhibitDescription')),
                    (dali1.emergency_rest.__name__, 'Rest', _('RestDescription')),
                )],
        }


