import json
import logging
import os
import re
import signal
import threading

from channels import layers
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse
from django.shortcuts import render, redirect

from reface import const
from reface.core import errors
from reface.dashboard import forms
from reface.dashboard.models import Box, Boxes, Controller
from resh.protocols.dali_utils.external_commands.dali2 import events
from resh.utils import errors

from reface.dashboard import fields
from resh.utils import const_values as v
from resh.utils.settings import conf

FIRMWARE_DIR = '/tmp/rapida/firmwares/'
BUILDS_DIR = '/tmp/rapida/builds/'
SETTINGS_DIR = '/var/project/reface.json'

logger = logging.getLogger('reface')


@login_required
def root(request):
    if not Boxes.instance.configured:
        return render(request, 'tools/configurator.html')
    if Boxes.instance.is_sw_gateway:
        return redirect('/rapida/0')
    return render(request, 'objects/system.html', context=dict(name=Boxes.instance.name,
                                                               boxes=sorted(Boxes.instance.values(),
                                                                            key=lambda box: box.id)))


@login_required
def upgrade(request):
    if request.method == 'POST':
        has_dependencies = False
        build = handle_uploaded_file(BUILDS_DIR, request.FILES['build'])
        for fs in request.FILES.getlist('dependencies', []):
            has_dependencies = True
            handle_uploaded_file(BUILDS_DIR, fs)

        cmd = f'sudo pip3 install {build} {"--break-system-packages" if Box.RBUS2 else "--upgrade"}'
        if has_dependencies:
            cmd += ' --no-index --find-links=%s' % BUILDS_DIR

        def update_and_exit(command):
            os.system(command)
            os.kill(os.getpid(), signal.SIGTERM)

        thr = threading.Thread(target=update_and_exit, args=[cmd])
        thr.start()
        return redirect_page(request, '/rapida/', 40)

    return render(request, 'tools/version_upgrade.html')


def redirect_page(request, url, timeout):
    return render(request, 'tools/redirector.html', context=dict(url=url, timeout=timeout))


@login_required
def configure(request):
    Boxes.instance.set_type(request.GET['type'])
    return redirect('/rapida/')


@login_required
def reload(request):
    Boxes.instance.init()
    return redirect('/rapida/')


@login_required
def controller(request):
    if 'update' in request.GET:
        if Boxes.instance.has_controller:
            Boxes.instance.controller.renew()
    elif 'disconnect' in request.GET:
        Boxes.instance.disconnect_sw_gateway()
        return redirect('/rapida/')
    elif 'nginx' in request.GET:
        threading.Thread(target=Controller.update_nginx).start()
        return redirect_page(request, request.GET.get('next', request.path), 15)
    elif 'upgrade' in request.GET:
        if 'time' in request.GET:
            Boxes.instance.controller.set_time(request.GET['time'])
        if 'date' in request.GET:
            Boxes.instance.controller.set_date(request.GET['date'])
        if 'timezone' in request.GET:
            Boxes.instance.controller.set_timezone(request.GET['timezone'])
        if 'dns0' in request.GET or 'dns1' in request.GET:
            Boxes.instance.controller.set_dns(
                ' '.join([request.GET.get('dns0', ''), request.GET.get('dns1', '')]).strip().split())
        if 'ip' in request.GET or 'mask' in request.GET or 'gateway' in request.GET:
            Boxes.instance.controller.set_ip(ip=request.GET.get('ip'), mask=request.GET.get('mask'),
                                             gateway=request.GET.get('gateway'))
    else:
        if Boxes.instance.has_controller:
            Boxes.instance.controller.renew()
    if request.GET:
        return redirect(request.path)

    return render(request, 'objects/controller.html', context=dict(name=Boxes.instance.name,
                                                                   is_sw_gateway=Boxes.instance.is_sw_gateway,
                                                                   controller=Boxes.instance.controller))


@login_required
def set_box_name(request, box_id):
    Boxes.instance.set_box_name(int(box_id), request.GET.get('name', 'unknown'))
    return redirect(request.GET.get('redirect', '/rapida/'))


@login_required
def set_name(request):
    Boxes.instance.set_name(request.GET.get('name', 'unknown'))
    return redirect(request.path[:request.path.index('set_name')])


@login_required
def add_box(request):
    occupied, index = Boxes.instance.get_indexes()
    action = request.GET.get('action', None)
    index = int(request.GET.get('index', index))
    host = re.sub(r'[\d+][:\d+]*$', '', request.get_host().split(':')[0])
    host = str(request.GET.get('host', host))
    port = int(request.GET.get('port', Box.DEF_PORT))
    name = request.GET.get('name', '')
    protocol=request.GET.get('protocol', 'R-BUS')
    if protocol == 'R-BUS2':
        protocol = Box.GATEWAY_RBUS2
    else:
        protocol = Box.GATEWAY_RBUS
    role = Box.FAKE
    protocols = ['R-BUS', 'R-BUS2']

    box = Box(index, host=host, port=port, name=name, role=role, protocol=protocol)

    if action == 'add':
        if is_address_occupied(host, port, occupied):
            return redirect(request.path)
        else:
            box.role = Box.GATEWAY
            Boxes.instance.set(box)
            return redirect(f'/rapida/box/{index}')
    elif action == 'conf':
        box.role = Box.GATEWAY
        Boxes.instance[index].status = Box.CONNECTED
        Boxes.instance.flush()
        return redirect(f'/rapida/box/{index}')
    elif action == 'remove':
        Boxes.instance.remove(index)
        return redirect('/rapida/')
    elif action == 'commit':
        Boxes.instance.set_box_name(index, name)
        return redirect(f'/rapida/box/{index}')

    return render(request, 'objects/box.html', context=dict(name=Boxes.instance.name, box=box,
                                                            occupied=occupied,
                                                            protocols=protocols))


def is_address_occupied(host: str, port: int, occupied: dict) -> bool:
    for addr in occupied:
        if str(addr['host']).lower() == host.lower() and int(addr['port']) == port:
            return True
    return False


@login_required
def box(request, box_id):
    _box = Boxes.instance.get(int(box_id))
    err = {}
    if 'update' in request.GET:
        Boxes.instance.update_box(_box)
        return redirect(request.path)
    elif 'readdress_modules' in request.POST:
        try:
            Boxes.instance.readdress(box_id)
        except BaseException as ex:
            err['readdress_modules'] = str(ex)
        else:
            Boxes.instance.update_box(box=_box)
            return redirect(request.path)
    protocols = ['R-BUS', 'R-BUS2']
    return render(request, 'objects/box.html', context=dict(
        name=Boxes.instance.name,
        box=_box,
        occupied=Boxes.instance.get_indexes(),
        err=err,
        protocols=protocols
    ))


@login_required
def channel(request, box_id, module_id, channel_id):
    try:
        _module = [m for m in Boxes.instance[int(box_id)].modules if m['address'] == int(module_id)][0]
    except IndexError:
        logging.error('module not found')
    else:
        return render(
            request, 'objects/channel.html',
            context=dict(
                name=Boxes.instance.name,
                box=int(box_id),
                box_name=Boxes.instance[int(box_id)].name,
                module=int(module_id),
                channel=int(channel_id),
                channels=(
                    ch['id']
                    for ch in _module.get('channels', [])
                    if ch.get('powered', False)
                )
            )
        )


@login_required
def line_builder(request, box_id, module_id, channel_id):
    form = forms.LineBuilder(*(request.POST,) if request.method == 'POST' else ())
    room = '%s/D/%s/%s' % (box_id, module_id, channel_id)
    if request.method == 'POST':
        if 'last_addressing' in request.POST:
            return redirect(request.path + 'process/')
        elif form.is_valid():
            try:
                command = form.command(request.POST.get('action'))
                try:
                    layers.get_channel_layer(const.LINE_BUILDER).build_line(
                        command, box_id, module_id, channel_id, form.mode, form.start_address
                    )
                except errors.AlreadyStarted:
                    pass
                return redirect(request.path + 'process/')
            except BaseException as ex:
                logging.error(ex)

    return render(request, 'tools/line_builder.html', context=dict(
        name=Boxes.instance.name, box=box_id, module=module_id, channel=channel_id, form=form, room=room,
        box_name=Boxes.instance[int(box_id)].name))


@login_required
def line_builder_process(request, box_id, module_id, channel_id):
    room = '%s/D/%s/%s' % (box_id, module_id, channel_id)
    return render(request, 'tools/line_building.html', context=dict(
        name=Boxes.instance.name, box=box_id, module=module_id, channel=channel_id, room=room,
        action_verbose=forms.LineBuilder.verbose_name(
            layers.get_channel_layer(const.LINE_BUILDER).action_name(box_id, module_id, channel_id)
        ),
        box_name=Boxes.instance[int(box_id)].name
    ))


@login_required
def ordering(request, box_id, module_id, channel_id, dali_type):
    room = '%s/D/%s/%s' % (box_id, module_id, channel_id)
    if dali_type in request.GET:
        addresses = [int(i[1:]) if i[1:].isdigit() else i[1:] for i in request.GET.get(dali_type).split(',') if len(i) > 0]
    else:
        addresses = range(0, 64)
    return render(
        request,
        'tools/ordering.html',
        context=dict(
            name=Boxes.instance.name, box=box_id, module=module_id, channel=channel_id, room=room,
            addr_range=range(0, 64),
            addresses=addresses,
            dali_type=dali_type,
            dali1_name=forms.dali1.Dali1Protocol.domain_name,
            dali2_name=forms.dali2.Dali2Protocol.domain_name,
            box_name=Boxes.instance[int(box_id)].name
        ))


@login_required
def probation(request, box_id, module_id, channel_id):
    room = '%s/D/%s/%s' % (box_id, module_id, channel_id)
    return render(
        request,
        'tools/probation.html',
        context=dict(
            name=Boxes.instance.name, box=box_id, module=module_id, channel=channel_id, room=room,
            addresses=range(0, 64),
            commands=(
                # Real command name, Displayed name
                (forms.dali1.gear.QueryActualLevel.__name__, forms.dali1.gear.QueryActualLevel.__name__),
                (forms.dali1.gear.QueryMaxLevel.__name__, forms.dali1.gear.QueryMaxLevel.__name__),
                (forms.dali1.gear.QueryMinLevel.__name__, forms.dali1.gear.QueryMinLevel.__name__),
                (forms.dali1.led.QueryDimmingCurve.__name__, forms.dali1.led.QueryDimmingCurve.__name__),
            ),
            box_name=Boxes.instance[int(box_id)].name
        ))


@login_required
def check_device(request, box_id, module_id, channel_id):
    room = '%s/D/%s/%s' % (box_id, module_id, channel_id)
    return render(
        request,
        'tools/check_device.html',
        context=dict(
            name=Boxes.instance.name, box=box_id, module=module_id, channel=channel_id, room=room,
            addresses=range(0, 64),
            commands=(
                # Real command name, Displayed name
                (forms.dali1.gear.QueryActualLevel.__name__, forms.dali1.gear.QueryActualLevel.__name__),
                (forms.dali1.gear.QueryMaxLevel.__name__, forms.dali1.gear.QueryMaxLevel.__name__),
                (forms.dali1.gear.QueryMinLevel.__name__, forms.dali1.gear.QueryMinLevel.__name__),
                (forms.dali1.led.QueryDimmingCurve.__name__, forms.dali1.led.QueryDimmingCurve.__name__),
            ),
            box_name=Boxes.instance[int(box_id)].name
        ))


@login_required
def devices(request, box_id, module_id, channel_id):
    form = forms.LineScanner(*(request.POST,) if request.method == 'POST' else ())
    settings = {
        'columns': [],
        'commands': []
    }
    if not os.path.exists(os.path.dirname(SETTINGS_DIR)):
        os.mkdir(os.path.dirname(SETTINGS_DIR))
    if request.method == 'POST':
        for item in request.POST:
            if not item.find("columns.") == -1:
                settings['columns'].append(item.rsplit('columns.', 1)[1])
            elif not item.find("commands.") == -1:
                settings['commands'].append(item.rsplit('commands.', 1)[1])
        settings = json.dumps(settings)
        with open(SETTINGS_DIR, 'w') as file:
            file.write(settings)
    try:
        with open(SETTINGS_DIR, 'r') as file:
            settings = json.loads(file.read())
    except BaseException as ex:
        settings['columns'] = [
            'fade_time',
            'actual_level',
            'event_scheme'
        ]
        settings['commands'] = [
            'discovery',
            'identify'
        ]
        logging.exception(ex)

    return render(request, 'tools/devices.html', context=dict(
        form=form,
        room='%s/D/%s/%s' % (box_id, module_id, channel_id),
        name=Boxes.instance.name, box=box_id, module=module_id, channel=channel_id,
        addresses=range(0, 64),
        groups_0_15=range(0, 16),
        groups_16_31=range(16, 32),
        columns=form.columns(),
        columns_width=6,
        commands=form.commands(),
        commands_width=5,
        event_filters=events.types,
        box_name=Boxes.instance[int(box_id)].name,
        settings=settings
    ))

@login_required
def fields_data(request, box_id, module_id, channel_id):
    # возвращает данные для селектов, имена полей с инпутами и чекбоксами
    data = json.dumps({
        'select': {
            v.RANGE_0_63: (fields.Field.of_range(0, 63, (255, 'MASK'))),
            v.RANGE_FADE_TIME: (fields.Field.of_fade_time()),
            v.RANGE_FADE_RATE: (fields.Field.of_fade_rate()),
            v.RANGE_DIMMING_CURVE: (fields.Field.of_dimming_curve()),
            v.RANGE_0_255: (fields.Field.of_range(0, 255)),
            v.RANGE_0_255_PERCENT: (fields.Field.of_range_with_percent(0, 255)),
            v.RANGE_63_255: (fields.Field.of_range(0, 63, (255, 'MASK'))),
            v.RANGE_0_25: (fields.Field.of_range(0, 25)),
            v.RANGE_2_5: (fields.Field.of_event_priority()),
            v.RANGE_EVENT_SCHEME: (fields.Field.of_event_scheme()),
            v.RANGE_EVENT_FILTER: (fields.Field.of_event_filter()),
            v.RANGE_OPERATING_MODE: (fields.Field.of_operating_mode()),
            v.RANGE_0_31_GROUP: (fields.Field.of_instance_group()),
            v.TIMER_10_0_254: (fields.Field.of_hold_time()),
            v.TIMER_05_0_255_2: (fields.Field.of_dead_time()),
            v.TIMER_1_0_255: (fields.Field.of_report_time()),
            v.TIMER_02_10_255_2: (fields.Field.of_short_time()),
            v.TIMER_02_10_100_2: (fields.Field.of_double_time()),
            v.TIMER_02_10_100_2_MIN: (fields.Field.of_double_time_min()),
            v.TIMER_02_5_100_2: (fields.Field.of_repeat_time()),
            v.TIMER_1_5_255: (fields.Field.of_stuck_time()),
        },
        'input': (v.CC_COOLEST, v.CC_WARMEST, v.CC_COLOR_TEMPERATURE, v.GROUPS, v.SCENES),
        'checkbox': (v.IDENTIFY, v.DISCOVERY, v.IS_ENABLED),
        'json': (v.EMERGENCY_MODE, v.EMERGENCY_STATUS, v.EMERGENCY_FEATURES, v.EMERGENCY_FAILURE_STATUS),
    })
    return HttpResponse(data)

@login_required
def module_dali(request, box_id, module_id):
    box_id = int(box_id)
    module_id = int(module_id)
    err = {}
    if 'upgrade' in request.POST:
        if 'address' in request.POST:
            address = int(request.POST.get('address'))
            if module_id != address:
                try:
                    Boxes.instance.set_module_address(box_id, module_id, address)
                except errors.DomainError as ex:
                    err['address'] = str(ex)
                else:
                    module_id = address
        if 'firmware' in request.FILES:
            firmware = handle_uploaded_file(FIRMWARE_DIR, request.FILES['firmware'])
            try:
                Boxes.instance.update_module_firmware(box_id, module_id, firmware)
            except errors.DomainError as ex:
                err['firmware'] = str(ex)
            else:
                return redirect(f'/rapida/{box_id}/D/{module_id}/')
            finally:
                os.remove(firmware)

        if not err:
            return redirect(f'/rapida/{box_id}/D/{module_id}/')
    return render(request, 'objects/module.html', context=dict(name=Boxes.instance.name,
                                                               box=box_id,
                                                               module=Boxes.instance.module_info(box_id, module_id),
                                                               modules=Boxes.instance.get_modules(box_id),
                                                               errors=err))


@login_required
def module_relay(request, box_id, module_id):
    box_id = int(box_id)
    module_id = int(module_id)
    err = {}
    if 'upgrade' in request.POST:
        if 'address' in request.POST:
            address = int(request.POST.get('address'))
            if module_id != address:
                try:
                    Boxes.instance.set_module_address(box_id, module_id, address)
                except errors.DomainError as ex:
                    err['address'] = str(ex)
                else:
                    module_id = address
        if 'firmware' in request.FILES:
            firmware = handle_uploaded_file(FIRMWARE_DIR, request.FILES['firmware'])
            try:
                Boxes.instance.update_module_firmware(box_id, module_id, firmware)
            except errors.DomainError as ex:
                err['firmware'] = str(ex)
            else:
                return redirect(f'/rapida/{box_id}/R/{module_id}/')
            finally:
                os.remove(firmware)
        if not err:
            return redirect(f'/rapida/{box_id}/R/{module_id}/')

    return render(request, 'objects/relay.html', context=dict(name=Boxes.instance.name,
                                                              box=box_id,
                                                              module=Boxes.instance.module_info(box_id, module_id),
                                                              modules=Boxes.instance.get_modules(box_id),
                                                              errors=err))


@login_required
def module_relay2(request, box_id, module_id):
    box_id = int(box_id)
    module_id = int(module_id)
    err = {}
    if 'upgrade' in request.POST:
        if 'address' in request.POST:
            address = int(request.POST.get('address'))
            if module_id != address:
                try:
                    Boxes.instance.set_module_address(box_id, module_id, address)
                except errors.DomainError as ex:
                    err['address'] = str(ex)
                else:
                    module_id = address
        if 'firmware' in request.FILES:
            firmware = handle_uploaded_file(FIRMWARE_DIR, request.FILES['firmware'])
            try:
                Boxes.instance.update_module_firmware(box_id, module_id, firmware)
            except errors.DomainError as ex:
                err['firmware'] = str(ex)
            else:
                return redirect(f'/rapida/{box_id}/R/{module_id}/')
            finally:
                os.remove(firmware)
        if not err:
            return redirect(f'/rapida/{box_id}/R/{module_id}/')

    if 'channel' in request.POST:
        channel_id = int(request.POST.get('channel'))
        state = 'off' if Boxes.instance.output_state(box_id, module_id, channel_id) else 'on'
        Boxes.instance.set_output_state(box_id, module_id, f'R{channel_id}', state)

        if not err:
            return redirect(f'/rapida/{box_id}/R/{module_id}/')

    return render(request, 'objects/relay2.html', context=dict(name=Boxes.instance.name,
                                                               box=box_id,
                                                               module=Boxes.instance.relay_info(box_id, module_id),
                                                               modules=Boxes.instance.get_modules(box_id),
                                                               errors=err))


@login_required
def relay_outputs(request, box_id, module_id):
    box_id = int(box_id)
    module_id = int(module_id)
    err = {}
    if 'upgrade' in request.POST:
        if 'R1' in request.POST:
            Boxes.instance.set_output_state(box_id, module_id, 'R1', request.POST['R1'])
        if 'R2' in request.POST:
            Boxes.instance.set_output_state(box_id, module_id, 'R2', request.POST['R2'])
        if 'R3' in request.POST:
            Boxes.instance.set_output_state(box_id, module_id, 'R3', request.POST['R3'])
        if 'R4' in request.POST:
            Boxes.instance.set_output_state(box_id, module_id, 'R4', request.POST['R4'])
        if 'R5' in request.POST:
            Boxes.instance.set_output_state(box_id, module_id, 'R5', request.POST['R5'])
        if 'R6' in request.POST:
            Boxes.instance.set_output_state(box_id, module_id, 'R6', request.POST['R6'])
        if 'R7' in request.POST:
            Boxes.instance.set_output_state(box_id, module_id, 'R7', request.POST['R7'])
        if 'R8' in request.POST:
            Boxes.instance.set_output_state(box_id, module_id, 'R8', request.POST['R8'])

        if not err:
            return redirect(f'/rapida/{box_id}/R/{module_id}/outputs/')

    return render(request, 'tools/outputs.html', context=dict(name=Boxes.instance.name,
                                                              box=box_id,
                                                              box_name=Boxes.instance[int(box_id)].name,
                                                              module=module_id,
                                                              outputs=Boxes.instance.output_states_by_mask(box_id, module_id),
                                                              errors=err))


@login_required
def relay_groups(request, box_id, module_id):
    box_id = int(box_id)
    module_id = int(module_id)
    err = {}

    if 'upgrade' in request.POST:
        if 'G1' in request.POST:
            Boxes.instance.set_group_state(box_id, module_id, 'G1', request.POST['G1'])
        if 'G2' in request.POST:
            Boxes.instance.set_group_state(box_id, module_id, 'G2', request.POST['G2'])
        if 'G3' in request.POST:
            Boxes.instance.set_group_state(box_id, module_id, 'G3', request.POST['G3'])
        if 'G4' in request.POST:
            Boxes.instance.set_group_state(box_id, module_id, 'G4', request.POST['G4'])
        if 'G5' in request.POST:
            Boxes.instance.set_group_state(box_id, module_id, 'G5', request.POST['G5'])
        if 'G6' in request.POST:
            Boxes.instance.set_group_state(box_id, module_id, 'G6', request.POST['G6'])
        if 'G7' in request.POST:
            Boxes.instance.set_group_state(box_id, module_id, 'G7', request.POST['G7'])
        if 'G8' in request.POST:
            Boxes.instance.set_group_state(box_id, module_id, 'G8', request.POST['G8'])

        if not err:
            return redirect(f'/rapida/{box_id}/R/{module_id}/groups/')

    return render(request, 'tools/groups.html', context=dict(name=Boxes.instance.name,
                                                             box=box_id,
                                                             box_name=Boxes.instance[int(box_id)].name,
                                                             module=module_id,
                                                             groups=Boxes.instance.group_states_by_mask(box_id,
                                                                                                        module_id),
                                                             errors=err))


def relay_config(request, box_id, module_id):
    box_id = int(box_id)
    module_id = int(module_id)
    err = {}

    return render(request, 'tools/relay_config.html', context=dict(name=Boxes.instance.name,
                                                                   box=box_id,
                                                                   box_name=Boxes.instance[int(box_id)].name,
                                                                   module=Boxes.instance.module_info(box_id, module_id),
                                                                   modules=Boxes.instance.get_modules(box_id),
                                                                   errors=err))


def group_config(request, box_id, module_id, group_id):
    box_id = int(box_id)
    module_id = int(module_id)
    group_id = int(group_id)
    err = {}

    return render(request, 'tools/group_config.html', context=dict(name=Boxes.instance.name,
                                                                   box=box_id,
                                                                   module=Boxes.instance.module_info(box_id, module_id),
                                                                   modules=Boxes.instance.get_modules(box_id),
                                                                   group_id=group_id,
                                                                   group_data=Boxes.instance.get_group_algo(box_id, module_id, group_id),
                                                                   errors=err))


def handle_uploaded_file(dest, f):
    os.makedirs(dest, exist_ok=True)
    filename = os.path.join(dest, f.name)
    with open(filename, 'wb+') as dest:
        for chunk in f.chunks():
            dest.write(chunk)
    return filename
