import ipaddress
import json
import os


from resh.domains._domain import Domain
from resh.templates.gateways import rgp_client_sv_config
from resh.utils import tools, compat
from resh.utils.const import GATEWAY, GATEWAY_IDs, GATEWAY_VAR, IP, PORT
from resh.utils.errors import BadParameter, BadGatewayIndex, StopExec

from resh.utils.gateway import gateways_status, get_files


class GatewayUtil:
    DEFAULT_PORT = 55577

    CAN_INTERFACE_DIR = '/etc/rapida/rcan/'
    CAN_INTERFACE_MASK = 'rcan'

    SUPERVISOR_CONF_EXT = '.conf'
    SPREAD_SERVICES_EXT = '.json'

    RGP_CLIENT_MASK = 'client_'
    RGP_SV_CLIENT_MASK = 'rgp_client_'

    SPREAD_SERVICES_CONF_DIR = '/etc/rapida/rgp/'
    SUPERVISOR_CONF_DIR = '/etc/supervisor/conf.d/'

    @staticmethod
    def generate_template(template_text: str, **replacement) -> str:
        for key, value in replacement.items():
            template_text = template_text.replace('{' + str(key) + '}', str(value))

        return template_text

    @classmethod
    def get_can_list(cls):
        """Список CAN-интерейсов на устройстве"""
        return [int(i) for i in get_files(cls.CAN_INTERFACE_DIR) if i.isdigit()]

    @classmethod
    def add_can(cls, gate_id: int):
        """ Добавить новый CAN-интерейс """
        can_file = os.path.join(cls.CAN_INTERFACE_DIR, str(gate_id))
        if not os.path.exists(can_file):
            tools.sys(f'sudo touch {os.path.join(cls.CAN_INTERFACE_DIR, str(gate_id))}')
        tools.sys(f'sudo service rcan reload')

    @classmethod
    def remove_can(cls, gate_id):
        """ Удалить существующий CAN-интерейс """
        can_file = os.path.join(cls.CAN_INTERFACE_DIR, str(gate_id))
        if os.path.exists(can_file):
            os.remove(can_file)
        tools.sys(f'sudo service rcan reload')

    @classmethod
    def add_sv_config(cls, filename: str, template: str, replacement={}):
        """ Добавить конфигурационный файл для автозапуска """
        template = GatewayUtil.generate_template(template, **replacement)
        tools.write_file(os.path.join(cls.SUPERVISOR_CONF_DIR, filename), template)
        tools.sys('sudo supervisorctl reread')
        tools.sys('sudo supervisorctl update')

    @classmethod
    def remove_sv_config(cls, file_name: str):
        """ Удалить конфигурационный файл для автозапуска шлюза """
        if file_name in get_files(cls.SUPERVISOR_CONF_DIR):
            os.remove(os.path.join(cls.SUPERVISOR_CONF_DIR, file_name))

        tools.sys('sudo supervisorctl reread')
        tools.sys('sudo supervisorctl update')


@compat.add_metaclass(compat.DomainTracker)
class GatewayDomain(Domain, GatewayUtil):
    def_params = {GATEWAY_VAR: 2}
    domain_name = GATEWAY

    PARAMETER_CAN_INTERFACE = 'canInterface'
    PARAMETER_HOST = 'remoteHost'
    PARAMETER_PORT = 'remotePort'
    RBUS_PROTOCOL = 'rbusVersion'

    DEFAULT_RGP_CONFIG = {
        "logLevel": 1,
        "remoteMinLatency": 100,
        "remoteMaxLatency": 10000,
        "remoteNAttempts": 10,
        "remoteKeepalive": 60
    }

    def exec(self): raise NotImplemented()

    @classmethod
    def get_rgp_list(cls):
        """ Список конфигурационных файлов для всех шлюзов """
        return get_files(cls.SPREAD_SERVICES_CONF_DIR, cls.RGP_CLIENT_MASK)

    @classmethod
    def add_rgp_config(cls, gate_id, ip, **kwargs):
        """ Добавить конфигурацию нового шлюза """
        try:
            host = str(ipaddress.ip_address(ip))
        except:
            raise BadParameter()

        template = {
            cls.PARAMETER_CAN_INTERFACE: f'{cls.CAN_INTERFACE_MASK}{gate_id}',
            cls.PARAMETER_HOST: host,
            cls.PARAMETER_PORT: cls.DEFAULT_PORT
        }
        template.update(GatewayDomain.DEFAULT_RGP_CONFIG)
        template.update(kwargs)

        file_name = os.path.join(
            cls.SPREAD_SERVICES_CONF_DIR,
            f'{cls.RGP_CLIENT_MASK}{gate_id}{cls.SPREAD_SERVICES_EXT}'
        )
        if os.path.exists(file_name):
            raise StopExec(f"Gateway<{gate_id}> already exist!")
        tools.write_file(file_name, json.dumps(template, indent=4))

    @classmethod
    def remove_rgp_config(cls, gate_id):
        """ Удалить конфигурацию шлюза """
        file_name = f'{cls.SPREAD_SERVICES_CONF_DIR}{cls.RGP_CLIENT_MASK}{gate_id}{cls.SPREAD_SERVICES_EXT}'
        if os.path.exists(file_name):
            os.remove(file_name)

    @property
    def id(self):
        try:
            return int(self._args[0])
        except:
            raise BadGatewayIndex()


class status(GatewayDomain):
    def_params = {}
    var_params = {GATEWAY_IDs: '1 2'}

    def exec(self):
        need_id = []
        try:
            while self.has_next:
                need_id.append(int(self.next_param))
        except:
            raise BadParameter()
        return gateways_status(need_id)



class connect(GatewayDomain):
    opt_params = {IP: '192.168.1.2'}
    var_params = dict(list({PORT: GatewayDomain.DEFAULT_PORT}.items()) + list(GatewayDomain.DEFAULT_RGP_CONFIG.items()))

    def exec(self):
        try:
            index = int(self.id)
            ip = str(ipaddress.ip_address(self.next_param))
        except:
            raise BadParameter()

        rgp_kwargs = {}

        if self.has_next and isinstance(self.next_param_recursive, int):
            rgp_kwargs[self.PARAMETER_PORT] = self.next_param

        if self.has_next and isinstance(self.next_param_recursive, int):
            protocol = self.next_param
        else:
            protocol = 0
        rgp_kwargs[self.RBUS_PROTOCOL] = protocol

        while self.has_next:
            pair = self.next_param
            try:
                key, value = pair.split('=')
                if key in self.DEFAULT_RGP_CONFIG:
                    rgp_kwargs[key] = value
            except BaseException:
                raise BadParameter(pair)

        replacements = dict()
        replacements['GATE_ID'] = index
        replacements['rbusVersion'] = protocol

        self.add_rgp_config(index, ip, **rgp_kwargs)
        self.add_can(index)
        self.add_sv_config(f'{self.RGP_SV_CLIENT_MASK}{index}{self.SUPERVISOR_CONF_EXT}', rgp_client_sv_config.template, replacements)

        return status(index).exec()[index]

    @classmethod
    def example(cls):
        return ' '.join([cls.domain_name, cls.__name__] +
                        [f'{cls.def_params[p]}' for p in cls.def_params] +
                        [f'{cls.opt_params[p]}' for p in cls.opt_params] +
                        [f'[{"{}=".format(p) if p != PORT else ""}{cls.var_params[p]}]' for p in cls.var_params])


class disconnect(GatewayDomain):
    def exec(self):
        index = self.id
        self.remove_sv_config(f'{self.RGP_SV_CLIENT_MASK}{index}{self.SUPERVISOR_CONF_EXT}')
        self.remove_rgp_config(index)
        self.remove_can(index)


class up(GatewayDomain):
    def exec(self): tools.sys(f'sudo supervisorctl start {self.RGP_SV_CLIENT_MASK}{self.id}')


class down(GatewayDomain):
    def exec(self): tools.sys(f'sudo supervisorctl stop {self.RGP_SV_CLIENT_MASK}{self.id}')


