import os

from resh.domains._domain import Domain
from resh.protocols.can_utils.rapida.const import CONTROLLER
from resh.protocols.can_utils.rapida import commands as rcommands, responses as rresp
from resh.protocols.can_utils.rapida2 import commands as r2commands
from resh.utils import groups, monitor_filter, tools, compat, settings
from resh.utils.const import *
from resh.utils.errors import CustomError, DeviceNotAnswered, ModuleConstructError, BadParameter, PacketParseError, \
    CanOpenError
from resh.utils.logger import log
from resh.domains.v1.module import Rapida1
from resh.domains.v2.module import Rapida2

COMMON_TYPE = 0
DALI_TYPE = 1
MODBUS_TYPE = 2
RELAY_TYPE = 4
DALI_WIRELESS_TYPE = 8

# Время для загрузки прошивки при перезапуске модуля, сек
REBOOT_DELAY = 5


@compat.add_metaclass(compat.DomainTracker)
class Module(Domain):
    def_params = {MODULE: 2}
    domain_name = MODULE_DOMAIN
    _module_type = 1
    dump_answer = False  # Временный признак что мне нужно все тело ответа а не только байты информации.

    def __init__(self, *args):
        super(Module, self).__init__(*args)
        global RapidaSend
        from resh.protocols.can_utils.rapida_packet import RapidaSend
        self._ext_mode = (settings.conf.protocol == 2)
        if settings.conf.device_type == 2:
            self.rapida_module = Rapida2
        elif settings.conf.device_type == 1:
            self.rapida_module = Rapida1
        else:
            raise CustomError(f'Unsupported device. Device type: {settings.conf.device_type}')

    def send_raw(self, frame: bytes):
        """Прослойка для работы update_firmware 1 и 2ой рапиды. По наличии времени уберу"""
        return self._send_raw(frame)

    def _send_raw(self, frame: bytes):
        if not isinstance(frame, (bytes, bytearray)):
            try:
                frame = bytes([int(b) for b in frame])
            except:
                raise BadParameter()

        message = RapidaSend(self._module_type, self.module, self._ext_mode, frame)
        return message.send_raw()

    def send_cmd(self, cmd, *params, filters=None, **kwargs):
        """Прослойка для работы update_firmware 1 и 2ой рапиды. По наличии времени уберу"""
        return self._send_cmd(cmd, *params, filters=filters, **kwargs)

    def _send_cmd(self, cmd, *params, filters=None, **kwargs) -> rresp.Response:
        if cmd == rcommands.Reboot or cmd == r2commands.R2Reboot:
            kwargs['answer_timeout'] = kwargs.get('answer_timeout', REBOOT_DELAY)

        try:
            _cmd = cmd(*[int(p) for p in params])
        except:
            raise BadParameter()

        log.info(f'-> {_cmd}')
        pack = RapidaSend(self._module_type, self.module, self._ext_mode, _cmd.frame, **kwargs)
        if _cmd.is_rbus2_proto:
            pack.arbitration_id_ext_bits = _cmd.ext_bits
        if filters is not None:
            pack.bus.set_filters(filters=filters)

        if self.dump_answer:
            pack.dump_answer = True
        return pack.send(_cmd.is_query)

    def _send(self, frame=None, need_answer=False, **kwargs):
        try:
            if frame is None:
                frame = []
                while self.has_next:
                    frame.append(int(self.next_param))

            elif isinstance(frame, int):
                frame = [frame]
            else:
                frame = list(frame)

            frame = bytes(frame)
        except:
            raise BadParameter()

        message = RapidaSend(self._module_type, self.module, self._ext_mode, frame, **kwargs)
        return message.send(need_answer=need_answer)

    def _listen_for(self, cmd_byte, timeout=0.05):
        message = RapidaSend(self._module_type, self.module, self._ext_mode, [cmd_byte - 1], answer_timeout=timeout)
        message.start_answer_timer()
        return message.receive()

    @property
    def module_type(self):
        return self._module_type

    @module_type.setter
    def module_type(self, value):
        self._module_type = value


"""https://kiwi.throne.systems/index.php/Tron/Strata/Hardware/Rapida/Resh/Command/Rapida#Base"""


class raw_command(Module, groups.Base):
    def_params = {CLASS: 1, MODULE_0_31: 2}
    opt_params = {BYTES_ARR: '1 0xff 255'}
    need_answer = False

    def exec(self):
        if not settings.conf.device_type == 1:
            raise CustomError('Unsupported command')
        try:
            self._module_type = int(self.next_param)
        except:
            raise BadParameter()
        self._send(need_answer=self.need_answer)


class raw_query(raw_command):
    need_answer = True


"""https://kiwi.throne.systems/index.php/Tron/Strata/Hardware/Rapida/Resh/Command/Rapida#Standard"""


class readdress(Module, groups.Common):
    def_params = {}
    var_params = {FORCED: ''}

    def exec(self):
        self._module = 0
        try:
            if self.ext_mode:
                self._send_cmd(r2commands.R2Reboot)
            else:
                self._send_cmd(rcommands.Reboot, answer_timeout=0.5)
        except DeviceNotAnswered:
            pass

        addressed = []
        forced = False

        if FORCED in self._args:
            forced = True
            self._args.remove(FORCED)

        try:
            self._send_cmd(rcommands.CommitAddresses, answer_timeout=5)
        except DeviceNotAnswered as ex:
            if forced:
                try:
                    self._send_cmd(rcommands.RunAutoAddressing, answer_timeout=5)
                except DeviceNotAnswered:
                    pass
                finally:
                    try:
                        self._send_cmd(rcommands.CommitAddresses)
                    except DeviceNotAnswered:
                        pass
            else:
                stack = ex.stack
                if stack is None or 0 not in stack:
                    raise ModuleConstructError('Expected answer from 01F is absent!')
                if len(stack) > 1:
                    stack.pop(0)
                    raise ModuleConstructError('Unexpected answer from modules {}'.format(stack.keys()))
                stack = stack.pop(0)
                unaddressed = []
                for dump in stack:
                    if isinstance(dump.payload.command, rresp.UnaddressedNotify):
                        if dump.data not in unaddressed:
                            unaddressed.append(dump.data)
                if len(unaddressed) == 0:
                    raise ModuleConstructError('no module has responded to 0xAD')

                expect_count = len(unaddressed)

                a = None
                try:
                    self._send_cmd(rcommands.RunAutoAddressing, filters=RapidaSend.filters(0x3E0), answer_timeout=5)
                except DeviceNotAnswered as ex:
                    a = ex
                    stack = ex.stack
                    if stack is None:
                        stack = {}
                    if CONTROLLER not in stack:
                        print('RunAutoAddressing(0xAA) command not answered')
                    else:
                        for dump in stack[CONTROLLER]:
                            if isinstance(dump.payload.command, rcommands.NextAddress):
                                addressed.append(dump.data[0] - 1)
                        if len(addressed) == 0:
                            print('No module answered on 0xAA')
                        elif addressed.count(1) > 1:
                            print('One of the not outer slots is empty')
                        elif expect_count != len(addressed):
                            print('{} modules found but {} addressed. Wtf?'.format(expect_count, len(addressed)))
                finally:
                    try:
                        self._send_cmd(rcommands.CommitAddresses)
                    except DeviceNotAnswered:
                        pass

                return len(addressed)


class find(Module, groups.Common):
    def_params = {}

    def __init__(self, *args):
        super(find, self).__init__(*args)
        global RapidaDump, ModuleDump
        from resh.protocols.can_utils.rapida_packet import RapidaDump, ModuleDump

    def exec(self):
        self._module = 0
        res = []
        try:
            if self.ext_mode:
                self._module_type = 0
                self._send_cmd(r2commands.R2QueryModuleType, filters=None)
            else:
                self._send_cmd(rcommands.QueryUidMin, filters=[])
        except DeviceNotAnswered as ex:
            if ex.stack:
                for m_id, dumps in ex.stack.copy().items():
                    for dump in dumps:
                        if isinstance(dump, (RapidaDump, ModuleDump)):
                            if self.ext_mode:
                                res.append(
                                    dict(address=dump.from_id,
                                         type=8 if dump.data.decode() == 'M08' else 1,
                                         name=dump.module_name,
                                         desc=dump.module_desc)
                                )
                            else:
                                res.append(
                                    dict(address=m_id,
                                         type=dump.module_type,
                                         name=dump.module_name,
                                         desc=dump.module_desc)
                                )
                            break
        return sorted(res, key=lambda item: item['address'])


class set_address(Module, groups.Common):
    opt_params = {'new_address': 4}

    def exec(self):
        new_address = self.next_param
        if self.ext_mode:
            _cmd = r2commands.R2SetAddress
            self._module_type = 0
            self._send_cmd(r2commands.R2RunProtectedCommand, 0x15)
        else:
            _cmd = rcommands.SetAddress

        self._send_cmd(_cmd, new_address)
        self._module = new_address
        self._listen_for(0x01, 5)


class info(Module, groups.Common):
    def exec(self):
        res = version(*self._args).exec()
        res['serial'] = serial(*self._args).exec()
        if self.ext_mode:
            res['interfaces'] = query_interface_types(*self._args).exec()
            res['type'] = get_type(*self._args).exec()
        return res


class _update_firmware(Module, groups.Common):
    opt_params = {FILE: 'path/to/file.ext'}
    """
    Пакеты, получаемые после обновления прошивки:
    BLFWReceived, BLFWSuccess, BootLoaderVersion, FirmwareVersion
    """
    def exec(self):
        path = self.next_param
        if not os.path.exists(path):
            raise CustomError('File "{}" not exist!'.format(path))
        if not os.path.isfile(path):
            raise CustomError('"{}" is not a file!'.format(path))
        return self.rapida_module.update_firmware(self, path, self.module)


class update_firmware(_update_firmware):
    pass


class update_power_firmware(_update_firmware):
    def_params = {}
    _module_type = 31
    _module = 30


class reboot(Module, groups.Common):
    def exec(self):
        if self.ext_mode:
            self._module_type = 0
            self._send_cmd(r2commands.R2Reboot)
        else:
            self._send_cmd(rcommands.Reboot)


class version(Module, groups.Common):
    def exec(self):
        if self._ext_mode is True:
            self._module_type = 0
            val = self._send_cmd(r2commands.R2QueryVersion).value
        else:
            val = self._send_cmd(rcommands.QueryVersion).value
        return dict(hardwareVersion=val['hv'],
                    firmwareVersion=val['fv'],
                    bootloaderVersion=val['bv'])


class get_type(Module, groups.Common):
    def exec(self):
        if self._ext_mode is True:
            self._module_type = 0
            return self._send_cmd(r2commands.R2QueryModuleType).params.decode()
        else:
            raise NotImplementedError()


class serial(Module, groups.Common):
    def exec(self):
        if self._ext_mode is True:
            self._module_type = 0
            return tools.byte_array_to_ascii_str(self._send_cmd(r2commands.R2QuerySerial).data, True)
        else:
            _serial = self._send_cmd(rcommands.QueryUidMax).data
            _serial += self._send_cmd(rcommands.QueryUidMin).data
        return ''.join(hex(b)[2:].rjust(2, '0').upper() for b in _serial)


class uid(Module, groups.Common):
    def exec(self):
        if self._ext_mode is True:
            self._module_type = 0
            _serial = self._send_cmd(r2commands.R2QueryUid, 0x02).data
            _serial += self._send_cmd(r2commands.R2QueryUid, 0x01).data
            _serial += self._send_cmd(r2commands.R2QueryUid, 0x00).data
        else:
            _serial = self._send_cmd(rcommands.QueryUidMax).data
            _serial += self._send_cmd(rcommands.QueryUidMin).data
        return ''.join(hex(b)[2:].rjust(2, '0').upper() for b in _serial)


"""
ALL MODULES COMMANDS
"""
class __R2Common(Module, groups.Common):
    _module_type = 0


class query_module_info(__R2Common):
    def exec(self):
        if self.ext_mode:
            _info = self._send_cmd(r2commands.R2QueryVersion).value
            return dict(hardwareVersion=_info['hv'],
                        firmwareVersion=_info['fv'],
                        bootloaderVersion=_info['bv'])
        else:
            raise NotImplementedError()


class query_serial(__R2Common):
    def exec(self):
        if self._ext_mode is True:
            self._module_type = 0
            _serial = self._send_cmd(r2commands.R2QuerySerial).value
            return dict(high=_serial['hb'],
                        low=_serial['lb'])
        else:
            raise NotImplementedError()


class query_interface_types(__R2Common):
    def exec(self):
        if self._ext_mode is True:
            _type = self._send_cmd(r2commands.R2QueryIfaceTypes)
            return ''.join(hex(b)[2:].rjust(2, '0').upper() for b in _type.data) + ': ' + _type.value
        else:
            raise NotImplementedError()


class query_controller_data(__R2Common):
    def exec(self):
        if self._ext_mode is True:
            _data = self._send_cmd(r2commands.R2QueryData).data
            return ''.join(hex(b)[2:].rjust(2, '0').upper() for b in _data)
        else:
            raise NotImplementedError()


class query_uid(__R2Common):
    """Запрос UID"""
    opt_params = {SECTION: 2}

    def exec(self):
        if self._ext_mode is True:
            if self.has_next:
                _section = self.next_param
                _uid = self._send_cmd(r2commands.R2QueryUid, _section).data
                return ''.join(hex(b)[2:].rjust(2, '0').upper() for b in _uid)
            else:
                raise BadParameter()
        else:
            raise NotImplementedError()


class query_module_state(__R2Common):
    """Запрос состояния модуля"""

    def exec(self):
        try:
            _state = self._send_cmd(r2commands.R2QueryModuleStatus).value
        except DeviceNotAnswered as ex:
            raise ex
        if _state:
            print(f'Причина перезагрузки модуля: {_state}')
        else:
            print('Причина перезагрузки модуля: неизвестно')


class silence_mode(__R2Common):
    """Режим тишины"""
    opt_params = {MODE: 0x02}

    def exec(self):
        if self._ext_mode is True:
            try:
                if self.has_next:
                    _mode = self.next_param
                    _state = self._send_cmd(r2commands.R2SilenceMode, _mode).params
                    if int(_mode) == 2:
                        if _state == 1:
                            print('Текущий режим тишины модуля: отправка в CAN отключена')
                        elif _state == 0:
                            print('Текущий режим тишины модуля: нормальная работа')
                        else:
                            print('Текущий режим тишины модуля: не определен!')
                else:
                    raise BadParameter('Обязательный параметр не указан!')
            except DeviceNotAnswered as ex:
                raise ex
        else:
            raise NotImplementedError()


class set_module_address(__R2Common):
    """Смена адреса модуля"""
    opt_params = {NEW_ADDRESS: 1}

    def exec(self):
        if self._ext_mode is True:
            try:
                if len(self.params) == 1:
                    _new_address = int(self.next_param)
                    if _new_address in range(0, 32):
                        self._send_cmd(r2commands.R2RunProtectedCommand, 0x15)
                        self._send_cmd(r2commands.R2SetAddress, _new_address)
                        self._module = _new_address
                        self._listen_for(0x01, 5)
                    else:
                        raise BadParameter('Адрес должен быть в диапазоне 0..31')
                else:
                    raise BadParameter('Новый адрес модуля не указан!')
            except DeviceNotAnswered as ex:
                raise ex
        else:
            raise NotImplementedError()


class set_module_config(__R2Common):
    """Смена скорости CAN модуля"""
    opt_params = {BAUD: 2400}

    def exec(self):
        try:
            if self.ext_mode is True:
                baud = rresp.ChannelConfig._values["baud"].index(self.next_param)
                return self._send_cmd(r2commands.R2SetChannelConfig, baud)
        except DeviceNotAnswered as ex:
            raise ex


class get_commands_list(__R2Common):
    def exec(self):
        if self._ext_mode is True:
            _cmd_list = self._send_cmd(r2commands.R2GetCommonCommandsList).value
            return print(f'{_cmd_list}', sep=";\n")
        else:
            raise NotImplementedError()


""" Dali """


class __Dali(Module, groups.Dali):
    _module_type = DALI_TYPE


class dali_status(__Dali):
    def exec(self):
        if self.ext_mode:
            r = self._send_cmd(r2commands.R2QueryChannelStatus).params
            r = bin(r)[:1:-1].ljust(8, '0')
        else:
            r = self._send_cmd(rcommands.QueryLineStatus)
            r = bin(r.data[0])[:1:-1].ljust(8, '0')
        result = [dict(id=i + 1, powered=r[i] == '1') for i in range(0, len(r))]
        return result


""" Modbus """


class __Modbus(Module, groups.Modbus):
    def_params = {MODULE: 1, CHANNEL: 2}
    _module_type = MODBUS_TYPE


class modbus_channel_status(__Modbus):
    def exec(self): return self._send_cmd(rcommands.QueryChannelConfig, self.channel).value


class modbus_reset_channel(__Modbus):
    def exec(self): return self._send_cmd(rcommands.ResetChannel, self.channel)


class modbus_set_channel_conf(__Modbus):
    opt_params = {BAUD: 2400, PARITY: 'odd', STOP_BITS: 2}

    def exec(self):
        try:
            baud = rresp.ChannelConfig._values["baud"].index(self.next_param)
            parity = rresp.ChannelConfig._values["parity"].index(self.next_param)
            stop = rresp.ChannelConfig._values["stop"].index(self.next_param)
        except:
            raise BadParameter()

        return self._send_cmd(rcommands.SetChannelConfig, self.channel, baud, parity, stop)


""" RELAY """


class __Relay(Module, groups.Relay):
    _module_type = RELAY_TYPE


class relay_outputs_status(__Relay):
    def exec(self):
        if self.ext_mode:
            r = self._send_cmd(r2commands.R2RelayQueryOutputState).value
            result = [dict(id=i + 1, powered=r.get(f'R{i + 1}')) for i in range(0, len(r))]
            return result
        else:
            raise NotImplementedError()


""" Wizards """


class monitor(Module, groups.Wizards):
    def_params = {}
    var_params = {MODULE: 1, CHANNEL: 1, DEVICE_ADDR: 'D3,-D4', INSTANCE: 'IT4', NAME: 'event'}

    def __init__(self, *args):
        super(monitor, self).__init__(*args)
        global CanConnection, CanDump
        from resh.protocols.can_utils.interface import CanConnection
        from resh.protocols.can_utils.can_packet import CanDump

    @classmethod
    def example(cls):
        return ' '.join([cls.domain_name, cls.__name__] +
                        [f'{cls.def_params[p]}' for p in cls.def_params] +
                        [f'{cls.opt_params[p]}' for p in cls.opt_params] +
                        [f'{cls.var_params[p]}' for p in cls.var_params]).lower() + \
            '\n' + cls.loc(cls.domain_name, cls.__name__, NAMED_PARAMS_DESC) + \
            '\n' + cls.loc(cls.domain_name, cls.__name__, EXCLUDE_PARAMS_DESC) + \
            '\n' + cls.loc(cls.domain_name, cls.__name__, BROADCAST_VALUE)

    def _prepare(self):
        self.filter = monitor_filter.Filter(self._args)
        self.can = CanConnection(interface=settings.conf.interface)

    def exec(self):
        try:
            self._prepare()
        except OSError as ex:
            raise CanOpenError(reason=str(ex))

        try:
            for dump in self.can.bus:
                try:
                    dump = CanDump.can_dump(dump)
                    if dump is None:
                        continue
                    payload = dump.payload
                    if self.filter.check(payload):
                        log.monitor(payload)
                except BaseException as ex:
                    try:
                        log.exception(PacketParseError(dump.payload.header_str() + ' ' + str(ex)))
                    except:
                        log.exception(PacketParseError(tools.bytes2str(dump)))
        except:
            return

    def set_break(self, sig, params):
        super(monitor, self).set_break(sig, params)
        self.can.stop()
