import resh.utils.tools
from resh.domains._domain import Domain
from resh.protocols.can_utils.rapida import responses as rresp
from resh.protocols.can_utils.rapida2 import commands as r2commands
#from resh.protocols.can_utils.relay_packet import RelaySend
from resh.utils import compat, groups, tools
from resh.utils.const import *
from resh.utils.errors import BadParameter, DeviceNotAnswered, NegativeResult
from resh.utils.logger import log


class RelaySettings:
    _group_id = 0xFF
    _add_group_frame = []
    _del_group_frame = []
    _count = 0
    _group_algo = {}

    def __init__(self):
        self.reset()

    def add_outputs_in_group(self, group: int, frame: list):
        self._group_id = group
        self._add_group_frame = frame

    def del_outputs_from_group(self, group: int, frame: list):
        if group == self._group_id:
            self._del_group_frame = frame
        else:
            raise BadParameter(f'Номер группы "{group + 1}" отличается от введенного ранее "{self._group_id}"!')

    def add_group_algo(self, params, frame):
        if self._count < 4:
            if self._count == 0:
                self._group_algo.update(algo0=dict(params=params, frame=frame))
            elif self._count == 1:
                self._group_algo.update(algo1=dict(params=params, frame=frame))
            elif self._count == 2:
                self._group_algo.update(algo2=dict(params=params, frame=frame))
            elif self._count == 3:
                self._group_algo.update(algo3=dict(params=params, frame=frame))
            self._count += 1
        else:
            raise BadParameter()

    @property
    def group_id(self) -> int:
        return self._group_id

    @property
    def add_group_frame(self) -> list:
        return self._add_group_frame

    @property
    def del_group_frame(self) -> list:
        return self._del_group_frame

    def step_algo_by_id(self, i: int) -> dict:
        if i == 0:
            return self._group_algo.get('algo0')
        elif i == 1:
            return self._group_algo.get('algo1')
        elif i == 2:
            return self._group_algo.get('algo2')
        elif i == 3:
            return self._group_algo.get('algo3')
        else:
            raise BadParameter()

    @property
    def steps(self) -> int:
        return self._count

    def reset(self):
        self._group_id = 0xFF
        self._add_group_frame.clear()
        self._del_group_frame.clear()
        self._count = 0
        self._group_algo.clear()


relay_settings = RelaySettings()


@compat.add_metaclass(compat.DomainTracker)
class RelayDomain(Domain):
    def_params = {MODULE: 4}
    domain_name = MODULE_RELAY
    _module_type = 4

    def __init__(self, *args):
        super(RelayDomain, self).__init__(*args)
        global RelaySend
        from resh.protocols.can_utils.relay_packet import RelaySend
        self._ext_mode = True

    # def _send_raw(self, frame: bytes):
    #
    #
    def _send_cmd(self, cmd, *params, filters=None, **kwargs) -> rresp.Response:
        try:
            _cmd = cmd(*[int(p) for p in params])
        except:
            raise BadParameter()

        ext_bits = _cmd.ext_bits if _cmd.is_rbus2_proto is True else 0

        log.info(f'-> {_cmd}')
        pack = RelaySend(self._module_type, self.module, _cmd.frame, cmd=1, **kwargs)
        pack.arbitration_id_ext_bits = ext_bits

        if filters is not None:
            pack.bus.set_filters(filters=filters)

        return pack.send(_cmd.is_query)

    def _send_msg(self, cmd, *params, filters=None, **kwargs) -> rresp.Response:
        try:
            _cmd = cmd(*[int(p) for p in params])
        except:
            raise BadParameter()

        ext_bits = _cmd.ext_bits if _cmd.is_rbus2_proto is True else 0

        log.info(f'-> {_cmd}')
        pack = RelaySend(self._module_type, self.module, _cmd.frame, cmd=0, **kwargs)
        pack.arbitration_id_ext_bits = ext_bits

        if filters is not None:
            pack.bus.set_filters(filters=filters)

        return pack.send(_cmd.is_query)

    # def _listen_for(self, cmd_byte, timeout=0.05):
    #     pass


"""
COMMANDS
"""


class query_info(RelayDomain, groups.Common):
    """ Запрос информации об интерфейсе """
    _module_type = 4

    def exec(self):
        try:
            _val = self._send_cmd(r2commands.R2RelayQueryInterface).value
        except DeviceNotAnswered as ex:
            raise ex
        return dict(relayMax=_val['rmax'],
                    groupMax=_val['gmax'])


class query_groups(RelayDomain, groups.Relay):
    """ Запрос сформированных групп """

    def exec(self):
        try:
            _groups = self._send_cmd(r2commands.R2RelayQueryGroups).value
        except DeviceNotAnswered as ex:
            raise ex
        if _groups:
            return _groups
        else:
            return 'Группы не сформированы'


class query_groups_algo(RelayDomain, groups.Relay):
    """ Запрос алгоритма переключения группы """
    opt_params = {GROUP_ID: 'G1', ALGO: 'ON/OFF', TIMEOUT: 'R/T'}
    var_params = {}

    def exec(self):
        try:
            if self.has_next:
                _group_id = resh.utils.tools.parse_group(self.next_param) - 1
            else:
                raise BadParameter('Номер группы не указан')

            # SW
            if self.has_next:
                _sw_str = str(self.next_param).lower()
                if _sw_str == 'on':
                    _param = 0x20
                elif _sw_str == 'off':
                    _param = 0
                else:
                    raise BadParameter('Режим группы указан некорректно')
            else:
                raise BadParameter('Режим группы не указан')

            # RT
            if self.has_next:
                _rt_str = str(self.next_param).lower()
                if _rt_str == 'r':
                    _param += 0x10
                elif _rt_str == 't':
                    _param = 0
                else:
                    raise BadParameter('Параметр R/T указан некорректно')
            else:
                raise BadParameter('Параметр R/T не указан')

            _algo = self._send_cmd(r2commands.R2RelayQueryGroupAlgo, _group_id, _param).value
        except DeviceNotAnswered as ex:
            raise ex
        return _algo


class _start_group_configuring_mode(RelayDomain, groups.AddConfig):
    """ Переход в режим конфигурирования группы
        (не должна вызываться напрямую)
    """
    opt_params = {GROUP_ID: 1}

    def exec(self):
        try:
            if len(self.params) != 1:
                raise BadParameter()
            else:
                _group_id = int(self.next_param)

            # run command "Protected command" first
            if r2commands.R2RelayConfigModeEntry.is_protected:
                _protected_command(*self._args).exec()

            _status = self._send_cmd(r2commands.R2RelayConfigModeEntry, _group_id)
            if _status.params != 0:
                raise BadParameter(f'Некорректный номер группы "{_status.value}" !')
        except DeviceNotAnswered as ex:
            raise ex
        return _status.value


class add_outputs_in_group(RelayDomain, groups.AddConfig):
    """Добавить выходы в группу"""
    opt_params = {GROUP_ID: 'G1'}
    var_params = {RELAY_NUMBER: 'R1,...R8'}

    def exec(self):
        try:
            if self.has_next:
                _group_id = tools.parse_group(self.next_param) - 1
            else:
                raise BadParameter('Номер группы не указан')

            if self.has_next:
                try:
                    _outputs = sorted(
                        [
                            tools.parse_relay(r) for r in self.next_param.split(',')
                        ]
                    )
                except:
                    raise BadParameter('Ошибка парсинга номеров реле')
            else:
                raise BadParameter('Ошибка парсинга номеров реле')

            relay_settings.add_outputs_in_group(_group_id, [tools.numbers2mask(_outputs)])
        except DeviceNotAnswered as ex:
            raise ex
        print(OK)


class del_outputs_from_group(RelayDomain, groups.AddConfig):
    """Удалить выходы из группы"""
    opt_params = {GROUP_ID: 'G1'}
    var_params = {RELAY_NUMBER: 'R1,...R8'}

    def exec(self):
        try:
            if self.has_next:
                _group_id = tools.parse_group(self.next_param) - 1
            else:
                raise BadParameter('Номер группы не указан')

            if self.has_next:
                try:
                    _outputs = sorted(
                        [
                            tools.parse_relay(r) for r in self.next_param.split(',')
                        ]
                    )
                except:
                    raise BadParameter('Ошибка парсинга номеров реле')
            else:
                raise BadParameter('Ошибка парсинга номеров реле')

            relay_settings.del_outputs_from_group(_group_id, [tools.numbers2mask(_outputs)])
        except DeviceNotAnswered as ex:
            raise ex
        print(OK)


class change_group_algo(RelayDomain, groups.AddConfig):
    """Изменение алгоритма переключения группы
        пример вызова: change_group_algo G1 ON R 1234, 5678, 1357, 2468
    """
    opt_params = {GROUP_ID: 'G1', ALGO: 'ON/OFF', TIMEOUT: 'R/T'}
    var_params = {STEP_DATA: '1234, 0, 5678'}

    def exec(self):
        try:
            # Group number
            if self.has_next:
                _group_id = tools.parse_group(self.next_param) - 1
            else:
                raise BadParameter('Номер группы не указан')

            # SW
            if self.has_next:
                _sw_str = str(self.next_param).lower()
                if _sw_str == 'on':
                    _param = 0x20
                elif _sw_str == 'off':
                    _param = 0
                else:
                    raise BadParameter('Режим группы указан некорректно')
            else:
                raise BadParameter('Режим группы не указан')

            # RT
            if self.has_next:
                _rt_str = str(self.next_param).lower()
                if _rt_str == 'r':
                    _param += 0x10
                elif _rt_str == 't':
                    _param += 0
                else:
                    raise BadParameter('Параметр R/T указан некорректно')
            else:
                raise BadParameter('Параметр R/T не указан')

            if self.has_next:
                try:
                    _steps_str = []
                    while self.has_next:
                        _step = str(self.next_param).split(',')
                        for _stp in _step:
                            if _stp:
                                _steps_str.append(_stp)
                except:
                    raise BadParameter('Ошибка парсинга параметров шагов')
            else:
                raise BadParameter('Ошибка парсинга параметров шагов')

            if len(_steps_str) not in range(0, 9):
                raise BadParameter('Кол-во шагов должно быть в пределах 0..8')

            try:
                _data = list()
                if _param & 0x10:
                    # маски реле
                    _data = [tools.numbers2mask(list(_relays)) for _relays in _steps_str]
                else:
                    # таймауты
                    _data = [int(_timeout) for _timeout in _steps_str]
            except:
                raise BadParameter('Ошибка парсинга параметров')

            relay_settings.add_group_algo(_param, _data)

            if relay_settings.steps == 4:
                save(*self.args).exec()

        except DeviceNotAnswered as ex:
            raise ex

        print(OK)


class save(RelayDomain, groups.AddConfig):
    """ Сохранение настроек группы """
    opt_params = {}
    var_params = {}

    def exec(self):
        # run command "Protected command" first
        _protected_command(*self._args).exec()

        # start group configuring mode
        _status = self._send_cmd(r2commands.R2RelayConfigModeEntry, relay_settings.group_id)
        if _status.params != 0:
            raise BadParameter(f'Некорректный номер группы "{_status.value}" !')

        # add outputs (relays) in group
        r2commands.R2RelayAddOutputInGroup.frame = relay_settings.add_group_frame
        self._send_cmd(r2commands.R2RelayAddOutputInGroup)

        # delete outputs if needed
        if len(relay_settings.del_group_frame) > 0:
            r2commands.R2RelayDelOutputFromGroup.frame = relay_settings.del_group_frame
            self._send_cmd(r2commands.R2RelayDelOutputFromGroup)

        # set algorithm
        for i in range(0, 4):
            _algo = relay_settings.step_algo_by_id(i)
            if _algo:
                r2commands.R2RelayChangeGroupAlgo.frame = _algo.get('frame')
                self._send_cmd(r2commands.R2RelayChangeGroupAlgo, _algo.get('params'))

        # save group
        _status = self._send_cmd(r2commands.R2RelaySaveConfig, relay_settings.group_id)
        if _status.params != 0:
            raise NegativeResult(f'Ошибка {_status.params} при сохренении группы {_status.value} !')
        else:
            print(OK)

        # clear settings
        relay_settings.reset()


class query_state_on_turning_on(RelayDomain, groups.TurningOn):
    """Запрос состояния реле и групп при включении"""
    opt_params = {}
    var_params = {}

    def exec(self):
        try:
            _state = self._send_cmd(r2commands.R2QueryRelayAndGroupsState).value
        except DeviceNotAnswered as ex:
            raise ex
        return dict(relayMax=_state['rmax'],
                    groupMax=_state['gmax'])


class set_state_on_turning_on(RelayDomain, groups.TurningOn):
    """Установка состояний при включении"""
    opt_params = {ALGO: 'R/G', MASK: 'M/N', NUMBER: 4, STATE: 'ON/OFF'}
    var_params = {}

    def exec(self):
        try:
            # R/G
            if self.has_next:
                _rg_str = str(self.next_param).lower()
                if _rg_str == 'g':
                    _param = 0x20
                elif _rg_str == 'r':
                    _param = 0
                else:
                    raise BadParameter('Режим группы указан некорректно')
            else:
                raise BadParameter('Режим группы не указан')

            # M/N
            if self.has_next:
                _mn_str = str(self.next_param).lower()
                if _mn_str == 'n':
                    _param += 0x10
                elif _mn_str == 'm':
                    _param += 0
                else:
                    raise BadParameter('Параметр M/N указан некорректно')
            else:
                raise BadParameter('Параметр M/N не указан')

            # NUM
            if self.has_next:
                _num = int(self.next_param)
                if _num in range(0, 9):
                    _param += _num
                else:
                    raise BadParameter('Номер реле/группы в диапазоне 0..8')
            else:
                raise BadParameter('Номер реле/группы не указан')

            # STATE/MASK
            if self.has_next:
                _state = str(self.next_param).lower()
                if _state == 'on':
                    _state = 1
                elif _state == 'off':
                    _state = 0
                else:
                    raise BadParameter('состояние реле/группы указано некорректно')
            else:
                raise BadParameter('состояние реле/группы не указано')

            # run command "Protected command" first
            _protected_command(*self._args).exec()
            # main command
            _cmd = r2commands.R2RelaySetStatesOnTurnOn
            _cmd.frame = [_state]
            _status = self._send_cmd(_cmd, _param)
            if _status.params != 0:
                raise NegativeResult(f'Ошибка "{_status.params}" при сохренении настроек!')
        except DeviceNotAnswered as ex:
            raise ex


class _protected_command(RelayDomain, groups.Common):
    """ Выполнение защищенной команды
        (не должна вызываться напрямую)
    """
    opt_params = {'new_param': 4}

    def exec(self):
        self._send_cmd(r2commands.R2RunProtectedCommand, 0x15)


class erase_settings(RelayDomain, groups.Common):
    """ Сброс всех настроек """

    def exec(self):
        try:
            if r2commands.R2RelayEraseSettings.is_protected:
                _protected_command(*self._args).exec()

            self._send_cmd(r2commands.R2RelayEraseSettings)
        except DeviceNotAnswered as ex:
            raise ex
        print(OK)


class get_commands_list(RelayDomain, groups.Common):
    """ Запрос поддерживаемых общих команд """

    def exec(self):
        try:
            _cmd_list = self._send_cmd(r2commands.R2GetCommonCommandsList).value
        except DeviceNotAnswered as ex:
            raise ex
        return print(f'{_cmd_list}', sep=";\n")


"""
DATA TRANSFER
"""


class query_output_state(RelayDomain, groups.RelayDataExchange):
    """Запрос состоянии выходов"""
    opt_params = {MASK: '0/1'}
    var_params = {NUMBER: 3}

    def exec(self):
        try:
            if len(self.params) < 1 or len(self.params) > 2:
                raise BadParameter()
            else:
                _params = list()
                _params.append(int(self.next_param))
                if self.has_next:
                    _params.append(int(self.next_param))
                _states = self._send_msg(r2commands.R2RelayQueryOutputState, *_params).value
        except DeviceNotAnswered as ex:
            raise ex
        return _states


class set_output_state(RelayDomain, groups.RelayDataExchange):
    """Установка состояния выходов"""
    opt_params = {MASK: '0/1', NUMBER: 3}
    var_params = {STATE: 1}

    def exec(self):
        try:
            if len(self.params) < 2:
                raise BadParameter()
            else:
                _params = list()
                _params.append(int(self.next_param))
                _params.append(int(self.next_param))
                if self.has_next:
                    _params.append(int(self.next_param))

                _error = self._send_msg(r2commands.R2RelaySetOutputState, *_params).value

                if _error:
                    print(f'Ошибка установки состояния выходов: "{_error}" !')
        except DeviceNotAnswered:
            pass
        except BadParameter:
            raise BadParameter('Ошибка ввода параметров')


class query_group_state(RelayDomain, groups.RelayDataExchange):
    """Запрос состоянии групп"""
    opt_params = {MASK: '0/1', NUMBER: 3}

    def exec(self):
        try:
            if len(self.params) < 2:
                raise BadParameter('Не указаны обязательные параметры')

            _num, _val = (int(self.next_param), int(self.next_param))
            _mn = 0
            if _num == 1:
                # number
                if _val not in range(1, 9):
                    raise BadParameter()
                else:
                    _mn = 1
                    _param = (_val << 5)
            elif _num == 0:
                # mask
                if _val not in range(1, 0x255):
                    raise BadParameter()
                else:
                    _param = _val
            else:
                # incorrect value
                raise BadParameter()

            _states = self._send_msg(r2commands.R2RelayQueryGroupState, _mn, _param).value
        except DeviceNotAnswered as ex:
            raise ex
        return _states


class set_group_state(RelayDomain, groups.RelayDataExchange):
    """Установка состояния групп"""
    opt_params = {MASK: '0/1', NUMBER: 3}
    var_params = {STATE: 1}

    def exec(self):
        try:
            if len(self.params) < 2:
                raise BadParameter()
            else:
                _params = list()
                _params.append(int(self.next_param))
                _params.append(int(self.next_param))
                if self.has_next:
                    _params.append(int(self.next_param))

            _error = self._send_msg(r2commands.R2RelaySetGroupState, *_params).value

            if _error:
                print(f'Ошибка установки состояния групп: "{_error}" !')
        except DeviceNotAnswered:
            pass


class drop_all_states_to_initial_state(RelayDomain, groups.RelayDataExchange):
    """Сброс всех состояний на состояния при включении"""
    opt_params = {}
    var_params = {}

    def exec(self):
        try:
            self._send_msg(r2commands.R2RelayDropAllStates)
        except DeviceNotAnswered as ex:
            raise ex


class query_common_commands_list(RelayDomain, groups.RelayDataExchange):
    """Запрос поддерживаемых общих команд"""
    opt_params = {}
    var_params = {}

    def exec(self):
        try:
            _list = self._send_msg(r2commands.R2RelayQuerySupportedCommonCommands).value
        except DeviceNotAnswered as ex:
            raise ex
        print(_list)
