import ipaddress
import os

import can

from resh.domains._domain import Domain
from resh.utils import groups, compat, settings
from resh.utils.const import SYSTEM, IP, DAY, YEAR, MONTH, MINUTES, HOURS, SECONDS, GATEWAY_VAR, \
    MASK_LC, DNS, DNS2, TZ
from resh.utils.errors import BadParameter, CustomError

DHCPCD = '/etc/dhcpcd.conf'


# System commands documentation
# https://kiwi.throne.systems/index.php/Tron/Strata/Hardware/Rapida/Resh/Command/System#System

@compat.add_metaclass(compat.DomainTracker)
class SystemDomain(Domain, groups.System):
    domain_name = SYSTEM

    def __init__(self, *args):
        super(SystemDomain, self).__init__(*args)
        if settings.conf.device_type == 2:
            self.storage_mnt = '/dev/mmcblk0p2'
        elif settings.conf.device_type == 1:
            self.storage_mnt = '/dev/root'
        else:
            raise CustomError(f'Unsupported device. Device type: {settings.conf.device_type}')


class get_device_type(SystemDomain):
    def exec(self): return settings.conf.device_type


class get_mac(SystemDomain):
    def exec(self): return os.popen('sudo ifconfig -a | grep ether').read().strip().split()[1].upper()


class get_ip(SystemDomain):
    def exec(self):
        try:
            _ip = os.popen(f'grep -i ^"static ip_address" {DHCPCD}').read().split('=')[1].strip()
            result = {GATEWAY_VAR: os.popen(f'grep -i ^"static routers=" {DHCPCD}').read().split('=')[1].strip()}
            if _ip:
                result[IP], mask = _ip.split('/')
                mask = ('1' * int(mask)).ljust(32, '0')
                result[MASK_LC] = '.'.join([f'{int(mask[i * 8:(i + 1) * 8], 2)}' for i in range(0, 4)])
            else:
                result[IP] = ''
            return result

        except Exception as e:
            print(f'Failed to get dns settings: {e}')
            raise e


class set_ip(SystemDomain):
    opt_params = {IP: '192.168.1.190'}
    var_params = {MASK_LC: '255.255.225.0', GATEWAY_VAR: '192.168.1.1'}

    def exec(self):
        try:
            try:
                ip, mask, gateway = str(ipaddress.ip_address(self.next_param)), '', ''
                if self.has_next:
                    mask = str(ipaddress.ip_address(self.next_param))
                    mask = ''.join(bin(int(b))[2:].rjust(8, '0') for b in mask.split('.')) + '0'
                    mask = len(mask[:mask.index('0')])
                    if self.has_next:
                        gateway = str(ipaddress.ip_address(self.next_param))
                if not mask:
                    mask = os.popen(f'grep -i ^"static ip_address" {DHCPCD}').read().split('=')[1].strip().split('/')[1]
                ip = f'{ip}/{mask}'
            except Exception as e:
                raise BadParameter(str(e))

            _set_net_param('ip_address', ip)
            if gateway:
                _set_net_param('routers', gateway)
            os.system(f'sudo ip addr flush dev {settings.conf.net_dev} && sudo systemctl restart dhcpcd.service')
        except BaseException as e:
            print(f'Failed to set ip settings: {e}')
            raise e


class get_dns(SystemDomain):
    def exec(self):
        try:
            return os.popen(f'grep -i ^"static domain_name_servers=" {DHCPCD}').read().strip().split('=')[1].split()
        except BaseException as e:
            print(f'Failed to get dns settings: {e}')
            raise e


class set_dns(SystemDomain):
    opt_params = {DNS: '4.4.4.4'}
    var_params = {DNS2: '8.8.8.8'}

    def exec(self):
        try:
            dns = []
            while self.has_next:
                dns.append(str(ipaddress.ip_address(self.next_param)))
            _set_net_param('domain_name_servers', ' '.join(dns))
        except BaseException as e:
            print(f'Failed to set dns settings: {e}')
            raise e


class get_date_time(SystemDomain):
    def exec(self):
        return {
            'datetime': os.popen('date --iso-8601=seconds').read().strip(),
            'timezone': os.popen('timedatectl | grep "Time zone"').read().split()[2]
        }


class set_date(SystemDomain):
    opt_params = {DAY: 31, MONTH: 12}
    var_params = {YEAR: 1970}

    def exec(self):
        try:
            day = str(self.next_param).rjust(2, '0')
            month = str(self.next_param).rjust(2, '0')
            try:
                year = str(self.next_param)
            except BadParameter:
                year = os.popen('date +%Y').read().strip()
        except:
            raise BadParameter()

        time = os.popen('date +%H:%M:%S').read().strip()
        date = f'{year}-{month}-{day}'

        res = os.popen(f'sudo date +%D -s "{time} {date}"').read().strip()
        if res and res != '{}/{}/{}'.format(month, day, year[-2:]):
            raise BadParameter(res)
        if os.popen('date +%Y-%m-%d').read().strip() != date:
            raise BadParameter('Unable to set date. Date/time synchronised.')
        os.popen('sudo hwclock --systohc')


class set_time(SystemDomain):
    opt_params = {HOURS: 23, MINUTES: 59}
    var_params = {SECONDS: 59}

    def exec(self):
        time = ':'.join([str(c).rjust(2, '0') for c in self.params])

        res = os.popen(f'date +%T -s "{time}"').read().strip()
        if res and res[:len(time)] != time:
            raise BadParameter(res)
        if os.popen('date +%H:%M:%S').read().strip() != time:
            raise CustomError('Unable to set time. Date/time synchronised.')
        os.popen('sudo hwclock --systohc')


class set_time_zone(SystemDomain):
    opt_params = {TZ: '+3'}

    def exec(self):
        try:
            time_zone_value = int(self.next_param)
            if 13 < time_zone_value < -15:
                raise
        except:
            raise BadParameter(f'Invalid time zone value')
        try:
            os.system(f'sudo timedatectl set-timezone Etc/GMT{"" if time_zone_value > 0 else "+"}{-time_zone_value}')
        except:
            raise BadParameter(f'Unable to change time zone')


class get_info(SystemDomain):
    def exec(self):
        return dict(model=os.popen('grep -i "Product: " /boot/rapida/device-info').read().strip().split('Product: ')[1],
                    imageVersion=os.popen('cat /boot/rapida/image-issue').read().strip())


class get_status(SystemDomain):
    def exec(self):
        res = {}
        try:
            cmd = f'df -h --output=source,size,used,avail,pcent | grep {self.storage_mnt}'
            flash_info = os.popen(cmd).read().split()
            res['flash'] = dict(total=flash_info[1], used=flash_info[2], free=flash_info[3])
        except:
            pass
        try:
            if settings.conf.device_type == 2:
                _temp = os.popen('cat /sys/class/hwmon/hwmon0/temp1_input').read().strip()
                res['temp'] = f"{_temp[:-3]}.{_temp[-3:-2]}'C"
            else:
                res['temp'] = os.popen('/opt/vc/bin/vcgencmd measure_temp').read().strip().split('=')[1]
        except:
            pass
        try:
            res['can'] = 'OK' if os.popen(f'sudo ifconfig {settings.conf.interface} | grep UP,RUNNING').read().strip() else 'not present'
        except:
            pass
        try:
            sd_size = int(os.popen('sudo blockdev --getsize64 /dev/mmcblk1').read().strip())
            res['sd'] = 'OK ({}G)'.format(round(sd_size/(1024**3), 1))
        except:
            res['sd'] = 'not present'

        return res


class can_dump(SystemDomain):
    def exec(self):
        from resh.protocols.can_utils.can_dump_handler import CanDumpHandler
        can_dump_handler = CanDumpHandler()
        notifire = can.Notifier(bus=can_dump_handler.bus, listeners=[can_dump_handler])
        while True:
            if input().lower().strip() == 'exit':
                break
        notifire.stop()


class reboot(SystemDomain):
    def exec(self): os.system('sudo reboot')


def _set_net_param(_name, _value):
    _static = f'static {_name}='
    last_value = os.popen(f'cat {DHCPCD} | grep "^{_static}"').read().strip()
    r = os.popen(f'sudo sed -i -e "s@^{last_value}@{_static}{_value}@g" {DHCPCD}').read().strip()
    if r:
        raise CustomError(r)
