import threading
from time import sleep

from resh.protocols.can_utils.can_packet import CanDump
from resh.protocols.can_utils.interface import CanConnection
from resh.protocols.can_utils.rapida_packet import RapidaDump
from resh.utils import settings
from resh.utils.errors import DeviceNotAnswered, NegativeResult
from resh.utils.logger import log
from resh.protocols.can_utils.rapida import commands, responses

TIMEOUT = 40


class Rapida1:

    @classmethod
    def update_firmware(cls, module_class, path, module_address):
        return UpdateFirmware(module_class, path, module_address).exec()

class UpdateFirmware:
    _stopped = False
    _can = None
    _exit_timer = None
    _exit_by_timeout = False
    """
    Пакеты, получаемые после обновления прошивки:
    BLFWReceived, BLFWSuccess, BootLoaderVersion, FirmwareVersion
    """
    def __init__(self,  module_class, path, module_address):
        self._path = path
        self._module_class = module_class
        self._module_address = module_address

    def exec(self):
        try:
            # start_quiescent_mode
            self._module_class.send_raw(bytes.fromhex('0102FFFFFE1D'))
            sleep(1)
            try:
                self._module_class.send_cmd(commands.Reboot, answer_timeout=0.05)
            except DeviceNotAnswered:
                pass
            self._module_class.send_cmd(commands.InitLoad, answer_timeout=5)
            self.exit_timer.start()
        except DeviceNotAnswered as ex:
            raise ex
        else:
            self._can = CanConnection(interface=settings.conf.interface)
            self._can.set_filter(self._module_address)
            threading.Thread(target=self.upload, args=[self._path]).start()
            result = None
            try:
                for dump in self._can.bus:
                    dump = CanDump.can_dump(dump)
                    if isinstance(dump, RapidaDump):
                        try:
                            payload = dump.payload
                            log.info('   <- %s' % payload.command)
                            if isinstance(payload.command, responses.BLFWSuccess):
                                result = True
                            elif isinstance(payload.command, responses.FirmwareVersion):
                                if result is None:
                                    raise NegativeResult('Модуль перезагружен, но прошивка не установлена')
                                elif isinstance(result, NegativeResult):
                                    raise result
                                else:
                                    return payload.command.value
                            elif isinstance(payload.command, (responses.BLFWError, responses.RebootReason)):
                                result = NegativeResult(str(payload.command))
                        except NegativeResult as ex:
                            raise ex
                        except BaseException as ex:
                            continue
            except NegativeResult as ex:
                raise ex
            except:
                pass
        finally:
            self.do_exit()
            # stop_quiescent_mode
            self._module_class.send_raw(bytes.fromhex('0102FFFFFE1E'))
            if self._exit_by_timeout:
                raise NegativeResult('Модуль не ответил в течение %s сек' % TIMEOUT)
    @property
    def exit_timer(self):
        if self._exit_timer is None:
            self._exit_timer = threading.Timer(TIMEOUT, function=self.on_timeout)
        return self._exit_timer

    def on_timeout(self):
        self._exit_timer = None
        self._exit_by_timeout = True
        self.do_exit()

    def do_exit(self):
        self._stopped = True
        if self._can:
            self._can.stop()
            self._can = None
        if self._exit_timer:
            self._exit_timer.cancel()

    def upload(self, path):
        with open(path, 'rb') as file:
            file.seek(0, 2)
            log.info('uploading %skB ...' % round(file.tell() / 1024))
            file.seek(0, 0)
            while not self._stopped:
                chunk = file.read(8)
                if chunk:
                    self._module_class.send_raw(chunk)
                    continue
                else:
                    log.info('uploading success')
                    return
            log.info('uploading break')
