import threading
from time import sleep

import can

from resh.protocols.can_utils.rapida.const import CONTROLLER, TYPES
from resh.protocols.can_utils.rapida2.const import TYPES as RAPIDA2_TYPES
from resh.utils import const, tools
from resh.utils.errors import DeviceNotAnswered, CanNotCommitted, CanOpenError, PacketParseError
from resh.utils.logger import log
from resh.utils.settings import conf
from resh.utils.tools import bytes2str

ANY_TYPE = 0
DALI_TYPE = 1
MODBUS_TYPE = 2
RELAY_TYPE = 4


class CanPacket:
    pr = 0
    cm = 1
    c_to_m = 1
    part = 0
    rfu1 = 0
    rfu2 = 0
    dump_answer = False # Временный признак что мне нужно все тело ответа а не только байты информации.

    def __init__(self, from_id, to_id, module_type, ext_mode, end_frame, cmd):
        self.from_id = from_id
        self.to_id = to_id
        self.module_type = module_type
        self.end_frame = end_frame
        self.cmd = cmd
        self._ext_mode = ext_mode

    def __str__(self):
        return bytes2str(self.frame)

    def __repr__(self):
        return f'[{self.__class__.__name__}]: {self.__str__()}'

    @property
    def module_address(self):
        if self.ext_mode:
            if self.c_to_m == 1:
                return self.to_id
            else:
                return self.from_id
        else:
            if self.from_id == CONTROLLER:
                return self.to_id
            else:
                return self.from_id

    @property
    def channel(self):
        return None

    @property
    def arbitration_id_ext_bits(self) -> int:
        """ extended bits (0..12) in arbitration id """
        raise NotImplementedError()

    @property
    def arbitration_id(self) -> int:
        """ arbitration id, 11 bits for old protocol, 29 bits for RBUS2 """
        if self._ext_mode is True:
            """ Extended CAN arbitration (29 bits) """
            return (self.pr << 28) + \
                (self.cm << 27) + \
                (self.c_to_m << 26) + \
                (self.module_address << 21) + \
                (self.part << 20) + (self.rfu1 << 19) + (self.rfu2 << 18) + \
                (self.module_type << 13) + \
                (self.cmd << 12) + \
                self.arbitration_id_ext_bits
        else:
            """ Common arbitration (11 bits) """
            return (self.pr << 10) + (self.from_id << 5) + self.to_id

    @property
    def ext_mode(self) -> bool:
        """ True, если используется протокол RBUS2 """
        return self._ext_mode

    @property
    def module_desc(self):
        tp = self.module_type
        if self.ext_mode:
            if tp in RAPIDA2_TYPES:
                return RAPIDA2_TYPES[tp]['desc']
        elif tp in TYPES:
            return TYPES[tp]['desc']
        return f'Unknown module type "{hex(tp)}"'

    @property
    def module_name(self):
        tp = self.module_type
        if self.ext_mode:
            if tp in RAPIDA2_TYPES:
                return RAPIDA2_TYPES[tp]['name']
        elif tp in TYPES:
            return TYPES[tp]['name']
        return f'Unknown module type "{hex(tp)}"'

    @property
    def frame(self) -> bytearray:
        raise NotImplementedError()

    @property
    def data(self) -> bytearray:
        raise NotImplementedError()

    @property
    def params(self) -> int:
        raise NotImplementedError()

    @property
    def error_code(self) -> int:
        raise NotImplementedError()

    @property
    def payload(self):  # -> Payload:
        raise NotImplementedError()

    @staticmethod
    def str_can_message(msg):
        return ('{}  {}   [{}]  {}'.format(conf.interface,
                                           hex(msg.arbitration_id)[2:].rjust(3, '0').upper(),
                                           len(msg.data),
                                           ' '.join(hex(b)[2:].rjust(2, '0').upper() for b in msg.data)).upper())


class CanSend(CanPacket):
    commit_timeout = 2
    answer_timeout = 0.5
    MAX_RETRY = 3

    @staticmethod
    def filters(arbitration_id: int) -> list:
        return [{"can_id": arbitration_id, "can_mask": 1023}]

    def __init__(self, module_type, module_address, ext_mode, end_frame, cmd=0, **kwargs):
        from_id = None if ext_mode else CONTROLLER
        super(CanSend, self).__init__(from_id, module_address, module_type, ext_mode, end_frame, cmd)
        self.bus_channel = conf.interface
        self.can_error = None
        self.answer_timer = None
        self.commit_timer = None
        self.need_state = 1
        self.need_answer = False
        self.listen_service = False
        self.state = 0
        self.stack = None
        self._bus = None
        self.send_twice = False
        self.by_channel_num = True  # Use channel number instead channel mask
        self.retry = 0  # Повтор отправки, если не пришел commit и retry < MAX_RETRY

        for key, value in kwargs.items():
            self.__setattr__(key, value)

    def start_answer_timer(self):
        if self.answer_timer:
            self.answer_timer.cancel()

        self.answer_timer = threading.Timer(self.answer_timeout, function=self.on_timeout, args=[DeviceNotAnswered])
        self.answer_timer.start()

    def stop_answer_timer(self):
        if self.answer_timer:
            self.answer_timer.cancel()
        self.answer_timer = None

    def start_commit_timer(self):
        self.stop_commit_timer()
        self.commit_timer = threading.Timer(self.commit_timeout, function=self.on_timeout, args=[CanNotCommitted])
        self.commit_timer.start()

    def stop_commit_timer(self):
        if self.commit_timer:
            self.commit_timer.cancel()
        self.commit_timer = None

    @property
    def _bus_filters(self) -> list:
        return [
            {
                "can_id": (self.module_address << 5) + CONTROLLER,
                "can_mask": 1023,
            },
            {
                "can_id": (1 << 27) + (self.module_address << 21),
                "can_mask": 0x8000000,  # 0xBFFFFFF
                "extended": True,
            },
        ]

    @property
    def bus(self):
        if self._bus is None:
            self._bus = can.interface.Bus(
                channel=self.bus_channel,
                bustype=const.BUS_TYPE,
                can_filters=self._bus_filters
            )

        return self._bus

    def on_timeout(self, err):
        self.can_error = err

    def check_commit(self, dump):
        raise NotImplementedError()

    @property
    def committed(self):
        return self.state == self.need_state

    def on_response(self, frame):
        raise NotImplementedError()

    def collect(self, a_id, dump):
        if self.stack is None:
            self.stack = dict()

        if self.ext_mode:
            mid = tools.get_bits(a_id, 21, 25)
        else:
            mid = (a_id >> 5) & 0x1F

        if mid not in self.stack:
            self.stack[mid] = []
        self.stack[mid].append(dump)

    def receive(self):
        while self.can_error is None:
            resp = self.bus.recv(0.05)
            if resp is not None:
                log.can(self.str_can_message(resp))
                dump = CanDump.can_dump(resp)
                if not dump or dump.ext_mode != self.ext_mode:
                    continue
                # Оригинальный пакет (адресация может быть как по каналу, так и по маске)
                # log.info('              {} {}'.format(hex(resp.arbitration_id)[2:].rjust(8, '0').upper(),
                #                                        ' '.join([f'{i:02X}' for i in resp.data]) if resp.data else ' '))

                # Распознанный пакет (адресация в нем всегда по каналу)
                log.info('   {} {} {} {}'.format(dump.__class__.__name__,
                                                 dump.channel if dump.channel else ' ',
                                                 hex(dump.arbitration_id)[2:].rjust(8, '0').upper(),
                                                 ' '.join([f'{i:02X}' for i in dump.data]) if dump.data else ' '))
                if self.to_id == 0:
                    self.collect(resp.arbitration_id, dump)
                elif self.to_id == dump.from_id:
                    if not self.committed:
                        self.check_commit(dump)
                        if self.committed:
                            if not self.need_answer:
                                log.info(f'CanSend return None ')
                                return
                            else:
                                log.info(f'   answer waiting')
                                self.start_answer_timer()
                    else:
                        result = self.on_response(dump)
                        if result is not None:
                            # log.info(f'CanSend return {result}')
                            return result if not self.dump_answer else dump

        log.info(f'{self.can_error}')
        raise self.can_error(stack=self.stack)

    def send(self, need_answer, listen_service=False):
        self.need_answer = need_answer
        self.listen_service = listen_service
        self.need_state = 2 if self.send_twice else self.need_state

        try:
            message = can.Message(arbitration_id=self.arbitration_id,
                                  is_extended_id=self.ext_mode,
                                  data=self.frame,
                                  extended_id=False)

            self.bus.send(message)
            log.can(self.str_can_message(message))

            if self.committed:
                log.info(f'   answer waiting')
                self.start_answer_timer()
            else:
                log.info(f'   commit waiting')
                self.start_commit_timer()

            if not self.committed or self.need_answer:
                try:
                    return self.receive()
                except CanNotCommitted as ex:
                    if self.retry < self.MAX_RETRY:  # Повтор отправки
                        self.retry += 1
                        self.can_error = None
                        sleep(0.1)
                        log.info('-> RETRY')
                        return self.send(self.need_answer, self.listen_service)
                    else:
                        log.info(f'CanSend raise {ex}')
                        raise ex
        except OSError as ex:
            raise CanOpenError(reason=str(ex))

    def send_raw(self):
        try:
            message = can.Message(arbitration_id=self.arbitration_id,
                                  is_extended_id=self.ext_mode,
                                  data=self.end_frame,
                                  extended_id=False)
            self.bus.send(message)
            log.can(self.str_can_message(message))
        except OSError as ex:
            raise CanOpenError(reason=str(ex))


class CanDump(CanPacket):
    def __init__(self, from_id, to_id, module_type, ext_mode, frame, cmd=0):
        super(CanDump, self).__init__(from_id, to_id, module_type, ext_mode, frame, cmd)
        self._arbitration_id_ext_bits = None
        self.c_to_m = 0

    @property
    def payload(self):
        raise NotImplementedError()

    @property
    def arbitration_id_ext_bits(self) -> int:
        return self._arbitration_id_ext_bits

    @arbitration_id_ext_bits.setter
    def arbitration_id_ext_bits(self, value):
        self._arbitration_id_ext_bits = value

    @property
    def frame(self):
        return self.end_frame

    @staticmethod
    def can_dump(message: can.Message) -> CanPacket:
        if 0x800 < message.arbitration_id < 0x20000000:
            return CanDump._can_dump_extended(message)
        else:
            return CanDump._can_dump_common(message)

    @staticmethod
    def _can_dump_common(message: can.Message) -> CanPacket:
        """Парсинг сообщения стандартного протокола"""
        from_id = (message.arbitration_id >> 5) & 0b11111
        to_id = message.arbitration_id & 0b11111
        frame = message.data
        b0 = frame[0]
        cmd = b0 >> 7
        cl = b0 & 0b11111    # class: 0 - тип, принимаемый всеми модулями, при условии CMD=1.
                             #        1	восьмиканальный модуль DALI
                             #        2	четырехканальный модуль Modbus
                             #        3	четырехканальный модуль DMX
                             #        4	восьмиканальный модуль реле

        if cmd == 0:
            if cl == 1:
                if from_id == CONTROLLER:
                    from .dali_packet import DaliSend
                    channel = DaliSend.channel_from_frame(frame)
                    return DaliSend(cl, to_id, channel, False, frame[3:],
                                    need_answer=DaliSend.need_answer_flag_from_frame(frame),
                                    send_twice=DaliSend.send_twice_flag_from_frame(frame))
                else:
                    """ CMD = 0 """
                    from .dali_packet import DaliDump
                    return DaliDump(from_id, to_id, cl, False, frame, cmd)
            elif cl == 2:
                if from_id == CONTROLLER:
                    from .modbusi_packet import ModbusSend
                    channel = ModbusSend.channel_from_frame(frame)
                    return ModbusSend(to_id, channel, False, frame[2:])
                else:
                    """ CMD = 0 """
                    from .modbusi_packet import ModbusDump
                    return ModbusDump(from_id, to_id, cl, False, frame, cmd)
        else:
            """ CMD = 1"""
            from .rapida_packet import RapidaDump
            return RapidaDump(from_id, to_id, cl, False, frame, cmd)

    @staticmethod
    def _can_dump_extended(message: can.Message) -> (CanPacket, None):
        """Парсинг сообщения протокола RBUS2"""
        cm = (message.arbitration_id >> 27) & 1  # 1 - обмен данными между контроллером и модулем, 0 - между модулем и модулем
        if cm != 1:
            return None
        c_to_m = (message.arbitration_id >> 26) & 1   # 1 - from controller to module, 0 - from module to controller
        module_id = tools.get_bits(message.arbitration_id, 21, 25)
        cmd = (message.arbitration_id & 0x1000) >> 12  # 1 - общая команда, 0 - команда передачи данных
        cmd_set = tools.get_bits(message.arbitration_id, 13, 17)  # номер набора команд модуля (module type)
        ext_bits = tools.get_bits(message.arbitration_id, 0, 11)

        if cmd == 0:
            if cmd_set == 1:  # DALI packet
                if c_to_m == 1:
                    from .dali_packet import DaliSend
                    channel = DaliSend.channel_from_arbitration_id(message.arbitration_id)
                    return DaliSend(cmd_set, module_id, channel, True, message.data,
                                    need_answer=DaliSend.need_answer_flag_from_arbitration_id(message.arbitration_id),
                                    send_twice=DaliSend.send_twice_flag_from_arbitration_id(message.arbitration_id))
                else:
                    from .dali_packet import DaliDump
                    _dump = DaliDump(module_id, None, cmd_set, True, message.data, cmd)
                    _dump.arbitration_id_ext_bits = ext_bits
                    return _dump
            # elif cl == 2:  # ModBus packet
            #     if from_id == CONTROLLER:
            #         from .modbusi_packet import ModbusSend
            #         channel = ModbusSend.channel_from_frame(frame)
            #         return ModbusSend(to_id, channel, True, frame[2:])
            #     else:
            #         from .modbusi_packet import ModbusDump
            #         return ModbusDump(from_id, to_id, cl, True, frame, cmd)
            # elif cl == 4:  # Relay packet
            #     if from_id == CONTROLLER:
            #         from .relay_packet import RelaySend
            #         _dump = RelaySend(cl, to_id, frame, cmd)
            #         _dump.arbitration_id_ext_bits = ext_bits
            #         return _dump
            #     else:
            #         from .relay_packet import RelayDump
            #         _dump = RelayDump(from_id, to_id, cl, True, frame, cmd)
            #         _dump.arbitration_id_ext_bits = ext_bits
            #         return _dump
        else:
            """ CMD = 1 (Общие команды) """
            if cmd_set == 0:  # ANY MODULE packet
                from .rapida_packet import ModuleDump
                _dump = ModuleDump(module_id, None, cmd_set, True, message.data, cmd)
                _dump.arbitration_id_ext_bits = ext_bits
                return _dump
            elif cmd_set == 1:  # DALI packet
                from .rapida_packet import RapidaDump
                _dump = RapidaDump(module_id, None, cmd_set, True, message.data, cmd)
                _dump.arbitration_id_ext_bits = ext_bits
                return _dump
            # elif cl == 4:  # Relay packet
            #     from .relay_packet import RelayDump
            #     _dump = RelayDump(from_id, to_id, cl, True, frame, cmd)
            #     _dump.arbitration_id_ext_bits = ext_bits
            #     return _dump
        # raise PacketParseError(CanPacket.str_can_message(message))

        return None
