from resh.protocols.can_utils.rapida.commands import Command
from resh.utils.errors import BadParameter

COMMON_TYPE = 0
DALI_TYPE = 1
RELAY_TYPE = 4


class R2Command(Command):
    _is_rbus2_proto = True
    _cmd_bit_pos = 12
    _com_cmd_bit_pos = 6
    _module_type = COMMON_TYPE

    @property
    def ext_bits(self) -> int:
        _params = 0
        for p in self.params:
            _params += int(p)
        return ((self._opcode if self._opcode is not None else 0) << self._com_cmd_bit_pos) + _params

    @property
    def frame(self) -> bytes:
        return bytes()

    @property
    def is_rbus2_proto(self):
        return self._is_rbus2_proto


class R2CommonQuery(R2Command):
    _query = True
    _com_cmd_bit_pos = 6


class R2DaliQuery(R2Command):
    _query = True
    _com_cmd_bit_pos = 8
    _module_type = DALI_TYPE


def of(opcode, module_type, frame) -> Command:
    for sc in R2Command.__subclasses__() + R2CommonQuery.__subclasses__():
        if issubclass(sc, R2Command):
            if (sc._opcode == opcode) and (sc._module_type == module_type):
                return sc(*frame)

    return R2Command(*frame)


"""
COMMON (ALL MODULES) COMMANDS
"""


class R2QueryVersion(R2CommonQuery):
    """ Запрос информации о модуле """
    _opcode = 0x00


class R2QuerySerial(R2CommonQuery):
    """ Запрос серийного номера модуля """
    _opcode = 0x01


class R2QueryIfaceTypes(R2CommonQuery):
    """ Запрос типов интерфейсов модуля """
    _opcode = 0x02


class R2QueryData(R2CommonQuery):
    """ Запрос данных микроконтроллера """
    _opcode = 0x03


class R2QueryUid(R2CommonQuery):
    """ Запрос UID """
    _opcode = 0x04

    @property
    def params(self):
        if int(self._params[0]) not in range(0, 3):
            raise BadParameter('mode must be in 0..2')
        return super(R2QueryUid, self).params


class R2QueryModuleStatus(R2CommonQuery):
    """ Запрос состояния модуля """
    _opcode = 0x05


class R2SilenceMode(R2CommonQuery):
    """ Режим тишины """
    _opcode = 0x06

    @property
    def params(self):
        if int(self._params[0]) not in range(0x0, 0x3):
            raise BadParameter('mode must be in 0..2')
        return super(R2SilenceMode, self).params


class R2QueryModuleType(R2CommonQuery):
    """ Запрос типа модуля """
    _opcode = 0x07


class R2SetAddress(R2CommonQuery):
    """ Смена адреса модуля """
    _opcode = 0x30
    _protected = True

    @property
    def params(self):
        if int(self._params[0]) not in range(0, 31):
            raise BadParameter('address must be in 1..29')
        return super(R2SetAddress, self).params


class R2SetChannelConfig(R2Command):
    """ Смена скорости CAN модуля """
    _opcode = 0x31

    def __init__(self, baud):
        """
        Настройка канала
        :param channel: номер канала (1-4)
        :param baud: скорость приема/передачи данных. Принимает значения:   0 - 1200
                                                                            1 - 2400
                                                                            2 - 4800
                                                                            3 - 9600
                                                                            4 - 19200
                                                                            5 - 38400
                                                                            6 - 57600
                                                                            7 - 115200
                                                                            8 - 230400
                                                                            9 - 460800
                                                                            10 - 921600


        :param parity - контроль четности. Принимает значения:  0 - none
                                                                1 - even
                                                                2 - odd
        :param stop - число стоп-битов. Принимает значения: 0 - 1 stop
                                                            1 - 2 stops
        """
        super(R2SetChannelConfig, self).__init__(*(baud))
        self.baud = baud


class R2RunProtectedCommand(R2Command):
    """Выполнение защищенной команды"""
    _opcode = 0x3D


class R2Reboot(R2CommonQuery):
    """Перезагрузка модуля"""
    _opcode = 0x3E


class R2GetCommonCommandsList(R2CommonQuery):
    """Запрос поддерживаемых общих команд"""
    _opcode = 0x3F


"""
DALI MODULE COMMANDS
"""


class R2QueryInterface(R2DaliQuery):
    _opcode = 0x00
    _module_type = DALI_TYPE


class R2QueryChannelStatus(R2DaliQuery):
    _opcode = 0x01
    _module_type = DALI_TYPE


"""
RELAY COMMANDS
"""


class R2RelayCommand(R2Command):
    _opcode = 0x00
    _module_type = RELAY_TYPE


class R2RelayQuery(R2CommonQuery):
    _opcode = 0x00
    _module_type = RELAY_TYPE


class R2RelayData(R2CommonQuery):
    _opcode = 0x00
    _module_type = RELAY_TYPE


def _get_relay_command(opcode, frame) -> R2RelayCommand:
    for sc in R2RelayCommand.__subclasses__():
        if issubclass(sc, R2RelayCommand):
            if sc._opcode == opcode:
                return sc(*frame)
    return R2RelayCommand(*frame)


def _get_relay_query(opcode, frame) -> R2RelayQuery:
    for sc in R2RelayQuery.__subclasses__():
        if issubclass(sc, R2RelayQuery):
            if sc._opcode == opcode:
                return sc(*frame)
    return R2RelayQuery(*frame)


def _get_relay_data_query(opcode, frame) -> R2RelayData:
    for sc in R2RelayData.__subclasses__():
        if issubclass(sc, R2RelayData):
            if sc._opcode == opcode:
                return sc(*frame)
    return R2RelayData(*frame)


def get_command(opcode, frame) -> R2Command:
    _cmd = _get_relay_command(opcode, frame)
    if _cmd.__class__ != R2RelayCommand:
        return _cmd
    else:
        _query = _get_relay_query(opcode, frame)
        if _query.__class__ != R2RelayQuery:
            return _query
        else:
            _data = _get_relay_data_query(opcode, frame)
            if _data.__class__ != R2RelayData:
                return _data
    return R2Command(*frame)


class R2RelayQueryInterface(R2RelayQuery):
    """ Запрос информации об интерфейсе """
    _opcode = 0x00


class R2RelayQueryGroups(R2RelayQuery):
    """Запрос сформированных групп"""
    _opcode = 0x10


class R2RelayQueryGroupAlgo(R2RelayQuery):
    """Запрос алгоритма переключения группы"""
    _opcode = 0x11

    @property
    def params(self):
        if int(self._params[0]) not in (0, 1, 2, 3, 4, 0x10, 0x20):
            raise BadParameter('mode must be in 0, 1, 2, 3, 4, 0x10, 0x20')
        return super().params


class R2RelayConfigModeEntry(R2RelayQuery):
    """Переход в режим конфигурирования группы"""
    _opcode = 0x15
    _protected = True

    # @property
    # def params(self):
    #     if int(self._params[0]) not in range(1, 8):
    #         raise BadParameter('Group ID must be in 1..8')
    #     return super(R2ConfigModeEntry, self).params()


class R2RelayAddOutputInGroup(R2RelayCommand):
    """Добавить выходы в группу"""
    _opcode = 0x16

    def __init__(self, *args):
        super().__init__(*args)
        self._frame = bytes()

    @property
    def frame(self) -> bytes:
        return self._frame

    @frame.setter
    def frame(self, value):
        self._frame = value


class R2RelayDelOutputFromGroup(R2RelayCommand):
    """Удалить выходы из группы"""
    _opcode = 0x17

    def __init__(self, *args):
        super().__init__(*args)
        self._frame = bytes()

    @property
    def frame(self) -> bytes:
        return self._frame

    @frame.setter
    def frame(self, value):
        self._frame = value


class R2RelayChangeGroupAlgo(R2RelayCommand):
    """Изменение алгоритма переключения группы"""
    _opcode = 0x18

    def __init__(self, *args):
        super().__init__(*args)
        self._frame = bytes()

    @property
    def frame(self) -> bytes:
        return self._frame

    @frame.setter
    def frame(self, value):
        self._frame = value


class R2RelaySaveConfig(R2RelayQuery):
    """Сохранение настроек группы"""
    _opcode = 0x19


class R2QueryRelayAndGroupsState(R2RelayQuery):
    """Запрос состояния реле и групп при включении"""
    _opcode = 0x20


class R2RelaySetStatesOnTurnOn(R2RelayQuery):
    """Установка состояний при включении"""
    _opcode = 0x21
    _protected = True

    def __init__(self, *args):
        super().__init__(*args)
        self._frame = None

    @property
    def frame(self) -> bytes:
        return self._frame

    @frame.setter
    def frame(self, value):
        self._frame = value


class R2RelayEraseSettings(R2RelayCommand):
    """Сброс всех настроек"""
    _opcode = 0x3E
    _protected = True


"""
RELAY DATA EXCHANGE MESSAGES
"""


class R2RelayQueryOutputState(R2RelayData):
    """Запрос состоянии выходов"""
    _opcode = 0x00
    _com_cmd_bit_pos = 9

    @property
    def params(self):
        if len(self._params) < 1:
            raise BadParameter()

        try:
            _num = int(self._params[0])
            if _num == 1:
                # number
                _val = int(self._params[1])
                if _val not in range(1, 9):
                    raise BadParameter()
                else:
                    return [(1 << 8) + ((_val - 1) << 5)]
            elif _num == 0:
                # mask
                return [0]
            else:
                # incorrect value
                raise BadParameter('Некорректный параметр!')
        except:
            raise BadParameter('Ошибка ввода параметров!')


class R2RelaySetOutputState(R2RelayData):
    """Установка состояния выходов"""
    _opcode = 0x01
    _com_cmd_bit_pos = 9

    @property
    def params(self):
        if len(self._params) < 2:
            raise BadParameter()

        try:
            _num, _val = (int(self._params[0]), int(self._params[1]))
            if _num == 1:
                # number
                if _val not in range(1, 9):
                    raise BadParameter()
                else:
                    _state = int(self._params[2])
                    return [(1 << 8) + ((_val - 1) << 5) + (_state << 3)]
            elif _num == 0:
                # mask
                if _val not in range(0, 0x255):
                    raise BadParameter()
                else:
                    return [_val]
            else:
                # incorrect value
                raise BadParameter('Некорректный параметр!')
        except:
            raise BadParameter('Ошибка ввода параметров!')


class R2RelayQueryGroupState(R2RelayData):
    """Запрос состоянии групп"""
    _opcode = 0x02
    _com_cmd_bit_pos = 9


class R2RelaySetGroupState(R2RelayData):
    """Установка состояния групп"""
    _opcode = 0x03
    _com_cmd_bit_pos = 9

    @property
    def params(self):
        if len(self._params) < 2:
            raise BadParameter('Не указаны обязательные параметры')

        try:
            _num, _val = (int(self._params[0]), int(self._params[1]))
            if _num == 1:
                # number
                if _val not in range(1, 9):
                    raise BadParameter()
                else:
                    _state = int(self._params[2])
                    if _state > 1:
                        raise BadParameter()
                    return [(1 << 8) + ((_val - 1) << 5) + (_state << 3)]
            elif _num == 0:
                # mask
                if _val not in range(1, 0x255):
                    raise BadParameter()
                else:
                    return [_val]
            else:
                # incorrect value
                raise BadParameter('Некорректный параметр!')
        except:
            raise BadParameter('Ошибка ввода параметров!')


class R2RelayDropAllStates(R2RelayData):
    """Сброс всех состояний на состояния при включении"""
    _opcode = 0x06
    _com_cmd_bit_pos = 9


class R2RelayQuerySupportedCommonCommands(R2RelayData):
    """Запрос поддерживаемых общих команд"""
    _opcode = 0x07
    _com_cmd_bit_pos = 9
