import time

import can

from resh.protocols.can_utils.rapida2.r_bus2_packet import FirmwarePacket, BootloaderPacket, packet_from_can_message
from resh.utils import const
from resh.utils.logger import log
from resh.utils.settings import conf


class FirmwareUpdateHandler(can.Listener):
	"""
	Can packages handler for firmware update event.
	"""
	# 0 - Waiting for flash start event, 1 - Uploading started, 2 - Uploading ended (uploading chunks)
	# 8 - Ended successfully,  9 - Ended with exception
	status = 0
	exception_text = ''

	def __init__(self, file_path, module_address):
		# TODO: Пока инициализация Can так, сделано что бы было единообразно с CanPacket. Передtлfть на общий CAN
		_bus_filters = [
				{
					"can_id": (module_address << 5) + 0x1f,
					"can_mask": 1023,
				},
				{
					"can_id": (1 << 27) + (module_address << 21),
					"can_mask": 0x8000000,  # 0xBFFFFFF
					"extended": True,
				}
			]
		self.bus = can.Bus(channel=conf.interface, bustype=const.BUS_TYPE, can_filters=_bus_filters)

		self.address = module_address
		self.file_path = file_path
		self.bootloader_packet = BootloaderPacket(module_address)
		self.firmware_packet = FirmwarePacket(module_address)

	def send(self, can_id, data):
		can_msg = can.Message(arbitration_id=can_id, data=data, is_extended_id=True)
		self.bus.send(can_msg)

	def send_firmware_chunk(self, chunk_number: int):
		self.firmware_packet.packet_number = chunk_number
		with open(self.file_path, 'rb') as file:
			file.seek(chunk_number * 8)
			chunk = file.read(8)
			if not chunk:
				return
			self.send(self.firmware_packet.packet_id(), chunk)

	def send_firmware(self):
		self.status = 1
		with open(self.file_path, 'rb') as file:
			file.seek(0, 2)
			log.info('uploading %skB ...' % round(file.tell() / 1024))
			file.seek(0, 0)

			self.firmware_packet.packet_number = 0
			chunk = file.read(8)
			while chunk:
				if self.firmware_packet.packet_number % 1000 == 0:
					self.send_command(0x00)
				self.send(self.firmware_packet.packet_id(), chunk)
				time.sleep(0.001)
				chunk = file.read(8)
				self.firmware_packet.increase_package_number()
		self.status = 2

	def send_command(self, command_param: int):
		self.bootloader_packet.param = command_param
		self.send(self.bootloader_packet.packet_id(), self.bootloader_packet.data)

	def on_message_received(self, msg: can.Message):
		packet = packet_from_can_message(msg)
		if packet.from_controller or not isinstance(packet, BootloaderPacket) or packet.address != self.address:
			return
		_param = packet.param

		if self.status == 0:
			# Module info after rebooting or waiting firmware in bootloader
			if _param == 0x01 or (_param == 0x0E and (packet.data[0] == 0x02 if packet.data else False)):
				self.send_command(0x10)
			# Module waiting for firmware packages
			elif _param == 0x10:
				self.send_firmware()
		# Module waiting resending firmware part by chunk number
		elif _param == 0x12:
			self.send_firmware_chunk(packet.data_value)
		# Module firmware successfully updated
		elif _param == 0x13:
			self.status = 8
			self.stop()
		# Module bootloader exception received
		elif _param == 0x0E:
			if packet.data_value != 0x05:
				self.status = 9
				self.exception_text = packet.data_text
				self.stop()
		elif _param == 0x00:
			pass
