import can

from resh.utils.errors import BadParameter


class RBus2Packet:
    """
    R-BUS2 Packet structure class
    By default this is packet from controller to module
    Packet format documentation:
    https://wiki.yandex.ru/electronics/razrabotka/programmnaja-chast/protokol-r-bus/format-paketa/
    """
    _pr = 0            # Can bus priority
    _c_m = 1           # 1 - Controller <-> Module, 0 - Module <-> Module
    _c_to_m = 1        # 1 - Controller -> Module, 0 - Controller <- Module  
    _part = 0          # 0 - No linked packages, 1 - one of linked packages
    _reserved = 0      # Reserved 19-18 bits
    _command_set = 0   # Command set number
    _cmd = 1           # Command type: 1 - To module, 2 - To interface< provide by module
    _command_data = 0  # Command information: number, params, ets. Depends on command_set
    packet_type = ''

    def __init__(self, address: int, **kwargs) -> None:
        self.address = address                  # Module address. 0 - broadcast message
        self.data = kwargs.get('data', bytes())  # Packet bytes
        
    def fill_with_binary(self, binary_string: str) -> None:
        self._pr = int(binary_string[-29:-28])
        self._c_m = int(binary_string[-28:-27])
        self._c_to_m = int(binary_string[-27:-26])
        self.address = int(binary_string[-26:-21], 2)
        self._part = int(binary_string[-21:-20])
        self._reserved = int(binary_string[-20:-18], 2)
        self._command_set = int(binary_string[-18:-13], 2)
        self._cmd = int(binary_string[-13:-12])
        self._command_data = int(binary_string[-12:], 2)

    def packet_id(self, type_=None):
        id_int = ((self._pr << 28) +
                  (self._c_m << 27) +
                  (self._c_to_m << 26) +
                  (self.address << 21) +
                  (self._part << 20) +
                  (self._reserved << 18) +
                  (self._command_set << 13) +
                  (self._cmd << 12) +
                  self._command_data)
        if type_ == hex:
            return f'{id_int:08X}'
        if type_ == bin:
            return f'{id_int:028b}'
        return id_int

    @property
    def from_controller(self) -> bool:
        return bool(self._c_to_m)

    @property
    def _custom_data_bin(self) -> str:
        """
        Returns custom 14 bits (19-18 reserved + 11-0) as 14 symbols bit string
        """
        return f'{self._reserved:02b}{self._command_data:012b}'

    def text_message(self, **kwargs) -> str:
        _bytes = ' '.join([f'{_byte:02X}' for _byte in self.data])
        _packet_id = f'{self.packet_id(hex)}#[{_bytes}]'
        _packet_id += ' ' * (34 - len(_packet_id))
        _packet_type = f'{self.packet_type or "Unknown"}'
        _packet_type = ' ' * (10 - len(_packet_type)) + _packet_type

        message = f'{_packet_id} {_packet_type} {"->" if self._c_to_m else "<-"} {self.address:2}'
        if kwargs.get('part'):
            message += f', Part: {self._part}'
        if not self.packet_type or kwargs.get('cmd'):
            message += f', CmdSet: {self._command_set}, Cmd: {self._cmd}'
        message += self.command_and_params_message(**kwargs)
        if not self.packet_type or kwargs.get('command_data'):
            message += self._command_and_params_bytes_message(self._reserved, self._command_data)
        message += self.comment_message()
        return message

    def command_and_params_message(self, **kwargs) -> str:
        return ''

    def comment_message(self, **kwargs) -> str:
        return ''

    @staticmethod
    def _command_and_params_bytes_message(_reserved: int, _command_data: int) -> str:
        return f', Reserved: {_reserved:02b}, CommandData: {_command_data:012b}'


class BootloaderPacket(RBus2Packet):
    """
    Bootloader packet use 0x3B (59) command in 11-6 bits and command param in 4-0 bits
    https://wiki.yandex.ru/electronics/razrabotka/programmnaja-chast/protokol-r-bus/nabory-komand/obshhie-komandy-1/
    """
    packet_type = 'Bootloader'

    __command_codes = {
        0x00: 'Keep alive',
        0x01: 'Информация о модуле',
        0x02: 'Серийный номер модуля',
        0x03: 'Данные микроконтроллера',
        0x04: 'UID_LOW микроконтроллера',
        0x05: 'UID_MID микроконтроллера',
        0x06: 'UID_HIGH микроконтроллера',
        0x07: 'Состояние модуля',
        0x08: 'Тип модуля',
        0x09: 'Смена адреса модуля',
        0x0A: 'Смена скорости CAN модуля',
        0x0B: 'Управление индикацией',
        0x0C: 'Установка адреса модуля по UID_LOW',
        0x0D: 'Переход на основную программу',
        0x0E: 'Ошибка',
        0x10: 'Готовность к обновлению',
        0x11: 'Получен пакет с последним номером',
        0x12: 'Запрос пакета по номеру',
        0x13: 'Обновление завершено успешно',
        0x1D: 'Доступные команды в текущем состоянии',
        0x1E: 'Перезагрузка модуля',
        0x1F: 'Поддерживаемые общие команды',
        0xFF: 'Unexpected command code'
    }

    __error_messages = {
        0x01: 'Отсутствует адрес модуля',
        0x02: 'Отсутствует основная программа',
        0x03: 'В модуль поступил пакет, формат которого не соответствует команде, указанной в пакете',
        0x04: 'Недопустимый адрес модуля',
        0x05: 'В модуль поступил пакет со Standard Identifier',
        0x10: 'Неверный формат заголовка файла',
        0x11: 'Слишком большой размер файла',
        0x12: 'Истек таймаут ожидания пакета',
        0x13: 'Ошибка записи в память',
        0x14: 'Ошибка контрольной суммы',
        0x15: 'Ошибка версии прошивки',
        0x16: 'Ошибка контрольной суммы перезаписанной прошивки',
        0x17: 'При поступлении  прошивки, дважды пришёл пакет с одинаковым номером',
        0x18: 'Слишком много пропущенных пакетов',
        0x19: 'Переполнение аппаратного буфера CAN',
        0xFF: 'Unexpected exception code'
    }

    def __init__(self, address: int, **kwargs) -> None:
        super().__init__(address, **kwargs)
        self._command_data = (0x3B << 6)

    @property
    def param(self) -> int:
        return self._command_data - 3776

    @param.setter
    def param(self, value: int):
        if value > 31:
            raise BadParameter(f'Bootloader parameter out of scope: {hex(value)}')
        self._command_data = (0x3B << 6) + (value << 0)
        self.data = bytes()

    @property
    def data_value(self):
        if self.param == 0x0E:
            return self.data[0] if self.data else 0xFF
        if self.param == 0x12:
            return (self.data[0] << 8) + (self.data[1])
        return ''

    @property
    def data_text(self):
        if self.param == 0x0E:
            return f'{self.data_value:02X}: {self.__error_messages.get(self.data_value, self.__error_messages[0xFF])}'
        return ''

    def command_and_params_message(self, **kwargs) -> str:
        return f', Command {self.param:02X}: {self.__command_codes.get(self.param, self.__command_codes[0xFF])}'

    def comment_message(self, **kwargs) -> str:
        return self.data_text


class GeneralPacket(RBus2Packet):
    packet_type = 'General'

    __command_codes = {
        0x00: 'Информация о модуле',
        0x01: 'Серийный номер модуля',
        0x02: 'Наборы команд модуля',
        0x03: 'Данные микроконтроллера',
        0x04: 'UID микроконтроллера',
        0x05: 'Состояние модуля',
        0x06: 'Режим тишины',
        0x07: 'Тип модуля',
        0x30: 'Смена адреса модуля',
        0x31: 'Смена скорости CAN модуля',
        0x3D: 'Выполнение защищенной команды',
        0x3E: 'Перезагрузка модуля',
        0x3F: 'Поддерживаемые общие команды',
        0xFF: 'Unexpected command code'
    }

    __error_messages = {
        0x01: 'На шине CAN обнаружен пакет с Stadart identifier. Байты DATA1, DATA0 содержат Stadart identifier полученного пакета',
        0x02: 'Переполнение аппаратного буфера CAN',
        0x03: 'Недопустимый адрес модуля (допустимые адреса: 0 … 31, 255)',
        0x04: 'Недопустимое значение скорости работы CAN',
        0xFF: 'Unexpected exception code'
    }

    @property
    def param(self) -> int:
        return int(self._custom_data_bin[8:14], 2)

    @param.setter
    def param(self, value: int):
        if value > 63:
            raise BadParameter(f'General param out of scope: {hex(value)}')
        self._command_data = (self.command << 6) + (value << 0)
        self.data = bytes()

    @property
    def command(self) -> int:
        return int(self._custom_data_bin[2:8], 2)

    @command.setter
    def command(self, value: int):
        if value > 63:
            raise BadParameter(f'General command out of scope: {hex(value)}')
        self._command_data = (value << 6) + (self.param << 0)
        self.data = bytes()

    def command_and_params_message(self, **kwargs) -> str:
        return (f', Command {self.command:02X}: {self.__command_codes.get(self.command, self.__command_codes[0xFF])}'
                f', Param: {self.param:02X}')

    def comment_message(self, **kwargs) -> str:
        if self.command == 0x0E:
            return f'{self.param:02X}: {self.__error_messages.get(self.param, self.__error_messages[0xFF])}'
        return ''


class DaliModulePacket(RBus2Packet):
    packet_type = 'DaliModule'

    __command_codes = {
        0x00: 'Информация об интерфейсе',
        0x01: 'Состояние каналов',
        0x02: 'Изменение состояния канала',
        0x0F: 'Поддерживаемые команды модуля DALI',
        0xFF: 'Unexpected command code'
    }

    def __init__(self, address: int, **kwargs) -> None:
        super().__init__(address, **kwargs)
        self._command_set = 1

    @property
    def param(self) -> int:
        return int(self._custom_data_bin[6:14], 2)

    @param.setter
    def param(self, value: int):
        if value > 63:
            raise BadParameter(f'General param out of scope: {hex(value)}')
        self._command_data = (self.command << 8) + (value << 0)
        self.data = bytes()

    @property
    def command(self) -> int:
        return int(self._custom_data_bin[2:6], 2)

    @command.setter
    def command(self, value: int):
        if value > 63:
            raise BadParameter(f'General command out of scope: {hex(value)}')
        self._command_data = (value << 8) + (self.param << 0)
        self.data = bytes()

    def command_and_params_message(self, **kwargs) -> str:
        param = None
        if self.command == 0:
            param = self.param
        elif self.command == 1:
            param = f'0b{self.param:08b}'
        elif self.command == 2:
            param = f'0b{self.param:05b}'
        return (f', Command {self.command:02X}: {self.__command_codes.get(self.command, self.__command_codes[0xFF])}'
                f'{", Param: " + param if param is not None else ""}')

    def comment_message(self, **kwargs) -> str:
        if self.command == 0x02:
            return (f', Канал {self.param - 16 if self.param > 15 else self.param}'
                    f' {"запитан" if self.param > 15 else "обесточен"}')
        return ''


class DaliBusPacket(RBus2Packet):
    packet_type = 'DaliBus'

    def __init__(self, address: int, **kwargs) -> None:
        super().__init__(address, **kwargs)
        self._command_set = 1
        self._cmd = 0

    @property
    def command_info(self) -> int:
        return int(self._custom_data_bin[6:14], 2)

    @property
    def answer(self) -> int:
        return int(self._custom_data_bin[3:4], 2)

    @answer.setter
    def answer(self, value: int):
        if value > 1:
            raise BadParameter(f'DaliBusPacket answer out of scope: {hex(value)}')
        self._command_data = (value << 10) + (self.double_send << 9) + (self.event_flag << 8) + (self.command_info << 0)
        self.data = bytes()

    @property
    def double_send(self) -> int:
        return int(self._custom_data_bin[4:5], 2)

    @double_send.setter
    def double_send(self, value: int):
        if value > 1:
            raise BadParameter(f'DaliBusPacket double_send out of scope: {hex(value)}')
        self._command_data = (self.answer << 10) + (value << 9) + (self.event_flag << 8) + (self.command_info << 0)
        self.data = bytes()

    @property
    def event_flag(self) -> int:
        """
        For packet to module 1 - 7-0 bits are chanel mask, 0 - 7-0 bits are chanel number
        For packet from module 1 - error
        """
        return int(self._custom_data_bin[4:5], 2)

    @event_flag.setter
    def event_flag(self, value: int):
        if value > 1:
            raise BadParameter(f'DaliBusPacket event_flag out of scope: {hex(value)}')
        self._command_data = (self.answer << 10) + (self.double_send << 9) + (value << 8) + (self.command_info << 0)
        self.data = bytes()

    def command_and_params_message(self, **kwargs) -> str:
        return f', Flags: {self.answer}{self.double_send}{self.event_flag}'


class DaliBusPacketToModule(DaliBusPacket):

    @property
    def channel(self) -> int:
        return self.command_info

    @channel.setter
    def channel(self, value):
        if value > 255:
            raise BadParameter(f'DaliBusPacketToModule channel out of scope: {hex(value)}')
        self._command_data = (self.answer << 10) + (self.double_send << 9) + (self.event_flag << 8) + (value << 0)
        self.data = bytes()

    def command_and_params_message(self, **kwargs) -> str:
        message = super().command_and_params_message(**kwargs)
        if self.event_flag:
            return f'{message}, Channel: {self.channel}'
        return f'{message}, Channel mask: {self.channel:08b}'


class DaliBusPacketFromModule(DaliBusPacket):

    __error_messages = {
        0x01: 'Нарушение длины бита принятого пакета',
        0x02: 'Нарушение стартовой последовательности',
        0x03: 'Некорректное число бит в пакете',
        0x04: 'Данные не отправлены в канал DALI. Переполнение очереди',
        0x05: 'Данные не отправлены в канал DALI. Канал DALI не запитан',
        0xFF: 'Unexpected exception code'
    }

    @property
    def error_code(self) -> int:
        return int(self._custom_data_bin[6:10], 2)

    @error_code.setter
    def error_code(self, value):
        if value > 15:
            raise BadParameter(f'DaliBusPacketFromModule error_code out of scope: {hex(value)}')
        self._command_data = (self.answer << 10) + (self.double_send << 9) + (self.event_flag << 8) + (value << 4) + (self.channel << 4)
        self.data = bytes()

    @property
    def channel(self) -> int:
        return int(self._custom_data_bin[10:14], 2)

    @channel.setter
    def channel(self, value):
        if value > 15:
            raise BadParameter(f'DaliBusPacketFromModule channel out of scope: {hex(value)}')
        self._command_data = (self.answer << 10) + (self.double_send << 9) + (self.event_flag << 8) + (self.error_code << 4) + (value << 4)
        self.data = bytes()

    def command_and_params_message(self, **kwargs) -> str:
        message = f'{super().command_and_params_message(**kwargs)}, Channel: {self.channel}'
        if self.event_flag:
            message += f'{self.error_code:02X}: {self.__error_messages.get(self.error_code, self.__error_messages[0xFF])}'
        return message


class FirmwarePacket(RBus2Packet):
    """
    Firmware packet use reserved 18 bit + 11-0 bits to keep firmware chunk number
    CMD (12 bit)
    """
    packet_type = 'Firmware'

    def __init__(self, address: int, packet_number: int = 0, **kwargs) -> None:
        super().__init__(address, **kwargs)
        self._cmd = 0
        self.packet_number = packet_number

    @property
    def packet_number(self) -> int:
        return self._command_data + (4096 * self._reserved)

    @packet_number.setter
    def packet_number(self, value: int):
        if value > 8191:
            raise BadParameter(f'Firmware packet number out of scope: {hex(value)}')
        self._reserved = 0 if value < 4096 else 1
        self._command_data = value if value < 4096 else value - 4096
        self.data = bytes()

    def increase_package_number(self) -> None:
        self.packet_number += 1
        self.data = bytes()

    def command_and_params_message(self, **kwargs) -> str:
        return f', Chunk number: {self.packet_number}'


def packet_from_can_message(msg: can.Message):
    """
    Returning Can packet instance from message, accounting message type.
    """
    bin_can_id = f'{msg.arbitration_id:029b}'
    packet = get_packet_class_by_can_id(bin_can_id)(0, data=msg.data)
    packet.fill_with_binary(bin_can_id)
    return packet


def get_packet_class_by_can_id(can_id):
    """
    Returning custom Can packet class by can message id.
    """
    set_ = int(can_id[-18:-13], 2)
    cmd = int(can_id[-13:-12])
    if set_ == 0x00 and cmd == 1:
        command_code = int(can_id[-12:-6], 2)
        return BootloaderPacket if command_code == 0x3B else GeneralPacket
    elif set_ == 0x00 and cmd == 0:
        return FirmwarePacket
    elif set_ == 0x01 and cmd == 1:
        return DaliModulePacket
    elif set_ == 0x01 and cmd == 0:
        direction = int(can_id[-27:-26])
        return DaliBusPacketToModule if direction else DaliBusPacketFromModule
    return RBus2Packet
