import resh.utils.tools
from resh.protocols.can_utils.rapida import responses
from resh.protocols.can_utils.rapida.responses import Response


class R2Response(Response):
    _is_rbus2_proto = True
    _module_type = 0

    def __init__(self, params, data=None):
        super().__init__(data)
        self._params = params

    @property
    def params(self) -> int:
        return self._params


def of(opcode, module_type, params, frame) -> Response:
    res = __iterate_subclasses(R2Response, opcode, module_type)
    if res:
        return res(params, frame)
    return R2Response(params, frame)


def __iterate_subclasses(cls, opcode: int, module_type: int):
    for sc in cls.__subclasses__():
        if (sc._opcode == opcode) and (sc._module_type == module_type):
            return sc
        elif len(sc.__subclasses__()) > 0:
            r = __iterate_subclasses(sc, opcode, module_type)
            if r:
                return r

    return None


class R2Version(R2Response):
    @property
    def value(self):
        ver = self.data

        def _decode(data: bytes):
            try:
                if data:
                    if data[0] == 0xFF:
                        return '--'
                    else:
                        return ''.join(hex(b)[2:].rjust(2, '0').upper() for b in data)
            except UnicodeDecodeError:
                return ''.join((hex(b)[2:].upper() for b in data))
        return {
            'hv': f'{chr(ver[0])}{ver[1]}',
            'bv': f'{chr(ver[2])}{ver[3]}',
            'fv': f'{chr(ver[4])}{ver[5]}'
        }


"""
MODULE COMMON RESPONSES
"""


class R2FirmwareVersion(R2Version):
    _opcode = 0x00


class R2Serial(R2Response):
    _opcode = 0x01

    @property
    def value(self):
        _ver = self.data

        def _decode(data: bytes):
            try:
                if data:
                    if data[0] == 0xFF:
                        return '--'
                    else:
                        return ''.join(
                            (hex(b)[2:].upper() for b in data))  # return '{}{}'.format(data[0:1].decode(), data[1])
            except UnicodeDecodeError:
                return ''.join((hex(b)[2:].upper() for b in data))

        return dict(lb=_decode(_ver[4:8]),
                    hb=_decode(_ver[0:4]))


class R2IfaceTypes(R2Response):
    _opcode = 0x02
    _values = {0x01: "Восьмиканальный модуль DALI",
               0x02: "Четырёхканальный модуль Modbus",
               0x03: "Четырёхканальный модуль DMX",
               0x04: "Восьмиканальный модуль реле",
               0x05: "Восьмиканальный модуль DO",
               0x06: "Восьмиканальный модуль дискретных входов",
               0x07: "Восьмиканальный модуль аналоговых входов", }

    @property
    def value(self):
        _type = self.data[0]
        return self._values.get(_type)


class R2Data(R2Response):
    _opcode = 0x03


class R2Uid(R2Response):
    _opcode = 0x04


class R2ModuleStatus(R2Response):
    _opcode = 0x05
    _values = {0: "PIN reset flag",
               1: "Power reset flag",
               2: "Software reset flag",
               3: "Independent watchdog reset flag",
               4: "Window watchdog reset flag",
               5: "Low-power reset flag",
               6: "Other"}

    @property
    def value(self):
        _p = self.params
        return self._values.get(_p)


class R2SilenceMode(R2Response):
    _opcode = 0x06


class R2Address(R2Response):
    _opcode = 0x30


class R2ChannelConfig(responses.Listable):
    _opcode = 0x31
    _values = dict(baud=[1200, 2400, 4800, 9600, 19200, 38400, 57600, 115200, 230400, 460800, 921600],
                   parity=['none', 'even', 'odd'],
                   stop=[1, 2])


class R2ProtectedCommand(R2Response):
    _opcode = 0x3D


class R2SupportedCommonCommands(R2Response):
    _opcode = 0x3F

    @property
    def value(self):
        def _decode(mask: int):
            _cmds = dict()
            for i in range(0, 8):
                if resh.utils.tools.get_bit(mask, i):
                    _cmds[f'cmd{i}'] = 'yes'
                else:
                    _cmds[f'cmd{i}'] = 'no'
            return _cmds

        _commands = dict()
        _bytes = self.data

        if _bytes and len(_bytes) > 0:
            for i in range(0, len(_bytes)):
                _list = _decode(_bytes[i])
                _commands[f'byte{i}'] = _list

        return _commands


"""
DALI MODULE RESPONSES
"""


class R2ServiceDali(R2Response):
    _module_type = 1
    _channel = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._channel = self.params


class R2ChannelStatus(R2ServiceDali):
    _opcode = 0x01

    @property
    def value(self):
        return f'mask: {bin(self.params)[2:].rjust(8, "0")}'


"""
RELAY MODULE RESPONSES
"""


class R2RelayResponse(R2Response):
    """Общий класс для ответов на команды модулю реле"""
    _module_type = 4


class R2RelayDataTransfer(R2Response):
    """Общий класс для ответов на передачу данных модулю"""
    _module_type = 4


def get_response(opcode, params, frame):
    res = __iterate_subclasses(R2RelayResponse, opcode, 4)
    if res:
        return res(params, frame)
    return R2RelayResponse(params, frame)


def get_data_response(opcode, params, frame):
    res = __iterate_subclasses(R2RelayDataTransfer, opcode, 4)
    if res:
        return res(params, frame)
    return R2RelayDataTransfer(params, frame)


class R2Interface(R2RelayResponse):
    """Запрос информации об интерфейсе"""
    _opcode = 0x00

    @property
    def value(self):
        _info = self.data

        def _decode(data: int):
            if data:
                if data == 0xFF:
                    return '--'
                else:
                    return f'{int(data)}'
            else:
                return '-'

        return dict(rmax=_decode(_info[0]),
                    gmax=_decode(_info[1]))


class R2QueryGroups(R2RelayResponse):
    """Запрос сформированных групп"""
    _opcode = 0x10

    @property
    def value(self):
        _groups = self.data

        def _decode(mask: int):
            _relays = list()
            if mask:
                if mask & 0x01:
                    _relays.append('R1')
                if mask & 0x02:
                    _relays.append('R2')
                if mask & 0x04:
                    _relays.append('R3')
                if mask & 0x08:
                    _relays.append('R4')
                if mask & 0x10:
                    _relays.append('R5')
                if mask & 0x20:
                    _relays.append('R6')
                if mask & 0x40:
                    _relays.append('R7')
                if mask & 0x80:
                    _relays.append('R8')
                return _relays
            else:
                return None

        def _add(_dest: str, _src: str):
            if _dest and len(_dest) > 0:
                _dest += ','
            _dest += _src
            return _dest

        if _groups and len(_groups) > 0:
            _result = dict()
            i = 1
            for _group in _groups:
                if _group:
                    _result[f'group{i}'] = _decode(_group)
                i += 1
            return _result
        else:
            return None


class R2QueryGroupAlgo(R2RelayResponse):
    """Запрос алгоритма переключения группы"""
    _opcode = 0x11

    @property
    def params(self) -> int:
        return self._params

    @property
    def value(self):
        _sw = self.params & 0x20
        _rt = self.params & 0x10
        if _rt:
            # relay masks
            _steps = self.data
            _masks = []
            for _s in _steps:
                _masks.append(dict(R1=bool(_s & 0x01),
                                   R2=bool(_s & 0x02),
                                   R3=bool(_s & 0x04),
                                   R4=bool(_s & 0x08),
                                   R5=bool(_s & 0x10),
                                   R6=bool(_s & 0x20),
                                   R7=bool(_s & 0x40),
                                   R8=bool(_s & 0x80), ))
            return _masks
        else:
            # timeouts
            _timeouts = {}
            for i in range(0, len(self.data)):
                _timeouts[f'step{i}'] = int(self.data[i])
            return _timeouts


class R2ConfigModeEntry(R2RelayResponse):
    """Переход в режим конфигурирования группы"""
    _opcode = 0x15

    @property
    def value(self):
        return self.data[0]


class R2SaveConfig(R2RelayResponse):
    """Сохранение настроек группы"""
    _opcode = 0x19


class R2RelayAndGroupsState(R2RelayResponse):
    """Запрос состояния реле и групп при включении"""
    _opcode = 0x20

    @property
    def value(self):
        _info = self.data

        def _decode(data: int):
            if data:
                if data == 0xFF:
                    return '--'
                else:
                    return f'{int(data)}'
            else:
                return '-'

        return dict(rmax=_decode(_info[0]),
                    gmax=_decode(_info[1]))


class R2SetStatesOnTurnOn(R2RelayResponse):
    """Установка состояний при включении"""
    _opcode = 0x21


class R2SupportedCommands(R2RelayResponse):
    """Запрос поддерживаемых общих команд"""
    _opcode = 0x3F

    @property
    def value(self):
        def _decode(mask: int):
            _cmds = dict()
            for i in range(0, 8):
                if resh.utils.tools.get_bit(mask, i):
                    _cmds[f'cmd{i}'] = 'yes'
                else:
                    _cmds[f'cmd{i}'] = 'no'
            return _cmds

        _commands = dict()
        _bytes = self.data

        if _bytes and len(_bytes) > 0:
            for i in range(0, len(_bytes)):
                _list = _decode(_bytes[i])
                _commands[f'byte{i}'] = _list

        return _commands


class R2QueryOutputState(R2RelayDataTransfer):
    """Состояние выходов"""
    _opcode = 0x00

    @property
    def value(self):
        _p = self.params
        _mn = resh.utils.tools.get_bit(_p, 8)
        if _mn == 1:
            # relay number
            _state = resh.utils.tools.get_bits(_p, 3, 4)
            _num = 'R' + str(resh.utils.tools.get_bits(_p, 5, 7) + 1)
            return dict([(_num, bool(_state))])
        else:
            # relays mask
            _state = resh.utils.tools.get_bits(_p, 0, 7)
            return dict(R1=bool(_state & 0x01),
                        R2=bool(_state & 0x02),
                        R3=bool(_state & 0x04),
                        R4=bool(_state & 0x08),
                        R5=bool(_state & 0x10),
                        R6=bool(_state & 0x20),
                        R7=bool(_state & 0x40),
                        R8=bool(_state & 0x80), )


class R2RelaySetOutputState(R2RelayDataTransfer):
    """Установка состояния выходов"""
    _opcode = 0x01

    @property
    def value(self):
        return str(self.params)


class R2RelayQueryGroupState(R2RelayDataTransfer):
    """Запрос состоянии групп"""
    _opcode = 0x02

    @property
    def value(self):
        _p = self.params
        _mn = resh.utils.tools.get_bit(_p, 8)
        if _mn == 1:
            # group number
            _state = resh.utils.tools.get_bits(self.params, 3, 4)
            _num = 'G' + str(resh.utils.tools.get_bits(_p, 5, 7) + 1)
            return dict([(_num, bool(_state))])
        else:
            # groups mask
            _state = resh.utils.tools.get_bits(_p, 0, 7)
            return dict(G1=bool(_state & 0x01),
                        G2=bool(_state & 0x02),
                        G3=bool(_state & 0x04),
                        G4=bool(_state & 0x08),
                        G5=bool(_state & 0x10),
                        G6=bool(_state & 0x20),
                        G7=bool(_state & 0x40),
                        G8=bool(_state & 0x80), )


class R2RelaySetGroupState(R2RelayDataTransfer):
    """Установка состояния групп"""
    _opcode = 0x03

    @property
    def value(self):
        return str(self.params)


class R2RelayQuerySupportedCommonCommands(R2RelayDataTransfer):
    """Запрос поддерживаемых общих команд"""
    _opcode = 0x07

    @property
    def value(self):
        return dict(func1=resh.utils.tools.get_bit(self.params, 0),
                    func2=resh.utils.tools.get_bit(self.params, 1),
                    func3=resh.utils.tools.get_bit(self.params, 2),
                    func4=resh.utils.tools.get_bit(self.params, 3),
                    func5=resh.utils.tools.get_bit(self.params, 4),
                    func6=resh.utils.tools.get_bit(self.params, 5),
                    func7=resh.utils.tools.get_bit(self.params, 6),
                    func8=resh.utils.tools.get_bit(self.params, 7), )
