from resh.utils.logger import log
from .can_packet import CanSend, CanDump
from .rapida import commands as rcmds
from .rapida import responses as rresp
from .rapida2 import commands as r2cmds
from .rapida2 import responses as r2resp
from ..payloads.rapida import ServicePacket
from ...utils import tools


class RapidaSend(CanSend):
    """ Command to DALI module """

    def __init__(self, module_type, module_address, ext_mode, end_frame, **kwargs):
        super(RapidaSend, self).__init__(module_type, module_address, ext_mode, end_frame, cmd=1, **kwargs)
        self._arbitration_id_ext_bits = 0

    def check_commit(self, dump):
        pass

    def on_response(self, dump):
        if isinstance(dump, (RapidaDump, ModuleDump)):
            log.info('   <- %s' % dump.payload.command)
            if self.ext_mode:
                self.stop_answer_timer()
                return dump.payload.command

            if (dump.command_byte == self.end_frame[0] + 1
                    or (dump.command_byte == rresp.FirmwareVersion._opcode
                        and self.end_frame[0] == rcmds.Reboot._opcode
                    )
            ):
                self.stop_answer_timer()
                return dump.payload.command

    @property
    def committed(self):
        return True

    @property
    def arbitration_id_ext_bits(self) -> int:
        return self._arbitration_id_ext_bits

    @arbitration_id_ext_bits.setter
    def arbitration_id_ext_bits(self, value):
        self._arbitration_id_ext_bits = value

    @property
    def frame(self):
        if self.ext_mode is True:
            return self.end_frame
        else:
            return ((1 << 7) + self.module_type).to_bytes(1, 'big') + self.end_frame


class RapidaDump(CanDump):
    """ Response for command in DALI module """
    cmd = 1

    @property
    def command_byte(self):
        if self.ext_mode:
            return tools.get_bits(self.arbitration_id_ext_bits, 8, 11)
        else:
            return self.frame[1]

    @property
    def data(self):
        if self.ext_mode:
            return self.frame
        else:
            return self.frame[2:]

    @property
    def params(self) -> int:
        if self.ext_mode:
            return tools.get_bits(self.arbitration_id_ext_bits, 0, 7)
        return 0

    @property
    def payload(self):
        if self.ext_mode:
            rs = r2resp.of(self.command_byte, self.module_type, self.params, self.data)
            if rs.__class__ == r2resp.R2Response:
                cm = r2cmds.of(self.command_byte, self.module_type, self.data)
                if cm.__class__ != r2cmds.R2Command and cm.__class__ != r2cmds.R2CommonQuery:
                    return ServicePacket(self, cm)
        else:
            rs = rresp.of(self.command_byte, self.data)
            if rs.__class__ == rresp.Response:
                cm = rcmds.of(self.command_byte, self.data)
                if cm.__class__ != rcmds.Command and cm.__class__ != rcmds.Query:
                    return ServicePacket(self, cm)
        return ServicePacket(self, rs)


class ModuleDump(CanDump):
    """ Response for command in any module """
    cmd = 1

    @property
    def command_byte(self):
        if self.ext_mode:
            return tools.get_bits(self.arbitration_id_ext_bits, 6, 11)
        else:
            return self.frame[1]

    @property
    def data(self):
        if self.ext_mode:
            return self.frame
        else:
            return self.frame[2:]

    @property
    def params(self) -> int:
        if self.ext_mode:
            return tools.get_bits(self.arbitration_id_ext_bits, 0, 5)
        return 0

    @property
    def payload(self):
        if self.ext_mode:
            rs = r2resp.of(self.command_byte, self.module_type, self.params, self.data)
            if rs.__class__ == r2resp.R2Response:
                cm = r2cmds.of(self.command_byte, self.module_type, self.data)
                if cm.__class__ != r2cmds.R2Command and cm.__class__ != r2cmds.R2CommonQuery:
                    return ServicePacket(self, cm)
            return ServicePacket(self, rs)
        else:
            return None
