import json
import signal
import sys
from datetime import datetime

from dali.command import BitmapResponse

from resh.utils.tools import StrByte

INFO = 1
CAN = 2
CAN_EXT = (3,)


class Log:
    def __init__(self):
        self.log_level = 0
        self.is_json = False

    def set_log_level(self, level: int):
        self.log_level = level

    def set_json_out(self, is_json: bool):
        self.is_json = is_json

    @staticmethod
    def _print_with_time(message):
        print('{}: {}'.format(datetime.now(), message))

    def _print_data(self, key, data, tab_count=0):
        if isinstance(data, dict):
            print(f'{"    "*tab_count}{key}: ')
            for key, value in data.items():
                self._print_data(key, value, tab_count+1)
        elif isinstance(data, (list, tuple)):
            if key:
                print(f'{"    "*tab_count}{key}:')
                tab_count += 1
            for i in data:
                print(f'{"    " * tab_count}{i}')
        else:
            print(f'{"    "*tab_count}{key}: {data}')

    def _easy_print(self, root_name: str, data):
        if self.is_json:
            print(json.dumps({root_name: data}, default=str))
        else:
            if isinstance(data, dict):
                for key, value in data.items():
                    self._print_data(key, value)
            elif isinstance(data, (list, tuple)) and len(str(data)) > 70:
                self._print_data('', data)
            elif data is not None:
                print(f'{data}')

    def result(self, data):
        if isinstance(data, BitmapResponse):
            d = dict()
            v = data._value[7:0]
            for b in data.bits:
                b_name = b.strip()
                while ' ' in b_name:
                    space = b_name.index(' ')
                    if len(b_name) - 1 > space:
                        b_name = b_name[:space] + b_name[space + 1].upper() + b_name[space + 2:]
                    else:
                        break

                d[b_name] = bool(v & 0x01 and b)
                v = (v >> 1)
            data = d
        elif isinstance(data, (bytes, bytearray)):
            if len(data) == 1:
                data = StrByte(data)
            else:
                data = ' '.join(hex(b)[2:].rjust(2, '0') for b in data)

        self._easy_print('result', data)

    def monitor(self, message):
        if self.is_json:
            message = message.to_dict()
        self._easy_print('can_message', message)

    def can(self, message):
        if CAN == self.log_level:
            self._print_with_time(message)
        if self.log_level in CAN_EXT:
            self._print_with_time(self.can_package_parser(message, self.log_level))

    def info(self, message):
        if INFO == self.log_level:
            self._print_with_time(message)

    def exception(self, ex: BaseException):
        if self.is_json:
            if hasattr(ex, 'serialize'):
                print(json.dumps(ex.serialize))
            else:
                print(json.dumps(str(ex)))
        else:
            print(str(ex).strip())

    def question(self, message, timeout=999):
        def alarm_handler(signum, frame):
            raise TimeoutError()
        signal.signal(signal.SIGALRM, alarm_handler)
        signal.alarm(timeout)
        try:
            return input(f'{message}>').strip()
        except:
            return None
        finally:
            sys.stdout.write('\r' + ' '*(len(message) + 1) + '\r')
            signal.alarm(0)

    @classmethod
    def can_package_parser(cls, message: str, type: int) -> str:
        """
        Парсинг сообщения из формата 28(32) бит в шестнадцатиричном формате
        К читаемому формату пакета, описанному в документации
        https://wiki.yandex.ru/electronics/razrabotka/programmnaja-chast/protokol-r-bus/format-paketa/
        C201102 -> ЭC_(T)_M: 11, ADDR: 0x1, PART: 0, SET: 0x0, CMD: 1, COM_CMD: 0x4, PRM: 0x2
        """
        try:
            _package = message.split()[1]
            _bits_package = bin(int(_package, 16))[2:].zfill(32)[::-1]
            _set = cls._get_bits(_bits_package, 13, 18, 'hex')
            _cmd = cls._get_bits(_bits_package, 12, 13)
            _reserved_bits = ''

            if _set == '0x0' and _cmd == "1":
                _com_cmd = cls._get_bits(_bits_package, 6, 12, "hex")
                _param_bits_end = 6
                if _com_cmd == '0x3B':
                    _type_msg = 'Команда загрузчика'
                    _param_bits_end = 5
                    _reserved_bits = f'{cls._standard_reserved_bits_data(_bits_package)}, 5: {_bits_package[5:6][::-1]}'
                else:
                    _type_msg = 'Общая команда модуля'
                    _param_bits_end = 5

                _data = f', COM_CMD: {_com_cmd}, PRM: {cls._get_bits(_bits_package, 0, _param_bits_end, "hex")}'
            elif _set == '0x0' and _cmd == "0":
                _bits = _bits_package[18:19] + _bits_package[0:12][::-1]
                _data = f', PAC_NUM: {int(_bits, 2)}'
            else:
                _data = f', DATA_COM: {cls._get_bits(_bits_package, 0, 12)}'

            _reserved_bits = _reserved_bits if _reserved_bits else cls._standard_reserved_bits_data(_bits_package)

            return (f'ЭC_(T)_M: {cls._get_bits(_bits_package, 26, 28)}'
                    f', ADDR: {cls._get_bits(_bits_package, 21, 26, "hex")}'
                    f', PART: {cls._get_bits(_bits_package, 20, 21)}'
                    f', SET: {_set}, CMD: {_cmd}{_data} ({_package}) {" ".join(message.split()[2:])}')
        except:
            return message

    @classmethod
    def _get_bits(cls, bin_string: str, start, end, base='bin'):
        _bits = bin_string[start:end][::-1]
        if base == 'hex':
            return hex(int(_bits, 2)).upper()
        elif base == 'int':
            return int(_bits, 2)
        return _bits

    @classmethod
    def _standard_reserved_bits_data(cls, bin_string: str) -> str:
        return f'PR: {bin_string[28:29]}, 19,18: {bin_string[18:20][::-1]}'


log = Log()
